import logging
import os
import pathlib as _pl
import pprint as pp
import typing as _tp
from datetime import datetime, timedelta

import numpy as np
import oemof.solph as solph
import pandas as pd
from oemof.tools import logger

import optihood.IO.writers as _wr
import optihood.entities as _ent
import optihood.IO.readers as read

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from optihood.buildings import Building
from optihood.constraints import *
from optihood._helpers import *
from optihood.links import Link
import optihood.IO.readers as _re

# TODO: define nr_of_buildings once  # pylint: disable=fixme


class OptimizationProperties:
    """
    Configurable attributes of energy networks.
    """
    def __init__(self,
                 optimization_type: _tp.Literal["costs", "env"],  # Pycharm highlights the input if anything else.
                 merge_link_buses: bool = False,
                 merge_buses: _tp.Optional[_tp.Sequence[str]] = None,
                 merge_heat_source_sink: bool = False,
                 temperature_levels: bool = False,
                 cluster_size: _tp.Optional[dict[str, int]] = None,
                 dispatch_mode: bool = False,
                 include_carbon_benefits: bool = False
                 ) -> None:
        """
        # TODO: explain inputs
        Parameters
        ----------

        cluster_size:
            Used to provide a selected number of days which could be assumed representative of the entire time range.

        optimization_type:
            "cost", or "env" depending on which criteria should be optimized.

        merge_link_buses:
            False: one bus of provided type per building.
            True: all buildings use the same bus for each type.

        merge_buses:
            Specify which buses to merge when merge_link_buses set to True

        merge_heat_source_sink:

        temperature_levels:

        dispatch_mode:
            True: activate dispatch optimization
            False: do investment + dispatch optimization

        include_carbon_benefits:

        """
        self.optimization_type = optimization_type
        self.merge_link_buses = merge_link_buses
        self.merge_buses = merge_buses
        self.merge_heat_source_sink = merge_heat_source_sink
        self.temperature_levels = temperature_levels
        self.cluster_size = cluster_size
        self.dispatch_mode = dispatch_mode
        self.include_carbon_benefits = include_carbon_benefits

        self.check_mutually_exclusive_inputs()

    def check_mutually_exclusive_inputs(self):
        if self.merge_link_buses and self.temperature_levels:
            logging.error("The options merge_link_buses and temperature_levels should not be set True at the same time."
                          "This use case is not supported in the present version of optihood. Only one of "
                          "mergeLinkBuses and temperatureLevels should be set to True.")


def get_data_from_df(df: pd.DataFrame, column_name: str):
    return df[column_name]


def get_data_from_excel_file(data: pd.ExcelFile, column_name: str):
    return data.parse(column_name)


class EnergyNetworkClass(solph.EnergySystem):
    def __init__(self, timestamp, clusters=None, temperatureLevels=False):
        self._timeIndexReal = timestamp             # the timeindex passed to the energy network, will be made shorter if clustering is selected
        self._clusterDate = {}
        if clusters:
            self._numberOfDays = len(clusters)
            lastDay = datetime(timestamp.year[0], 1, 1) + timedelta(self._numberOfDays - 1)
            lastDay = lastDay.strftime('%Y-%m-%d')
            i = 1
            for day in clusters:
                d_clusterIndex = datetime(timestamp.year[0], 1, 1) + timedelta(i - 1)
                self._clusterDate[day] = d_clusterIndex.strftime('%Y-%m-%d')
                i += 1
            timestamp = pd.date_range(f"{timestamp.year[0]}-01-01 00:00:00", f"{lastDay} 23:00:00", freq=timestamp.freq)
        self._nodesList = []
        self._thermalStorageList = []
        self._storageContentSH = {}
        self._storageContentDHW = {}
        self._storageContentTS = {}
        self._storageContentPIT = {}                # TODO: check and update the code for calculation of storage content for pit
        self._storageContent = {}
        self.__technology_efficiency = {}
        self.__inputs = {}                          # dictionary of list of inputs indexed by the building label
        self.__technologies = {}                    # dictionary of list of technologies indexed by the building label
        self.__capacitiesTransformersBuilding = {}  # dictionary of dictionary of optimized capacities of transformers indexed by the building label
        self.__capacitiesStoragesBuilding = {}      # dictionary of dictionary of optimized capacities of storages indexed by the building label
        self.__costParam = {}
        self.__envParam = {}
        self.__capex = {}
        self.__opex = {}
        self.__feedIn = {}
        self.__envImpactInputs = {}                 # dictionary of dictionary of environmental impact of different inputs indexed by the building label
        self.__envImpactTechnologies = {}           # dictionary of dictionary of environmental impact of different technologies indexed by the building label
        self._busDict = {}
        self.__elHP = {}
        self.__shHP = {}
        self.__dhwHP = {}
        self.__intermediateOpTempsHP = {}
        self.__annualCopHP = {}
        self.__elGWHP = {}
        self.__shGWHP = {}
        self.__dhwGWHP = {}
        self.__intermediateOpTempsGWHP = {}
        self.__annualCopGWHP = {}
        self.__elRodEff = np.nan
        self._temperatureLevels = temperatureLevels
        self._dispatchMode = False
        self._c = 4.186 #default value for water in kJ/(kg K)
        self._rho = 1. #default value for water in kg/L
        _pl.Path(".\\log_files").mkdir(parents=True, exist_ok=True)
        screen_format = "%(asctime)s-%(levelname)s:  %(message)s"
        logger.define_logging(screen_format=screen_format, logpath=os.getcwd(), logfile=f'.\\log_files\\optihood_{datetime.now().strftime("%d.%m.%Y %H.%M.%S")}.log')
        logging.info("Initializing the energy network")
        super(EnergyNetworkClass, self).__init__(timeindex=timestamp, infer_last_interval=True)

    def setFromExcel(self, filePath, numberOfBuildings, clusterSize={}, opt="costs", mergeLinkBuses=False, mergeBuses=None, mergeHeatSourceSink=False, dispatchMode=False, includeCarbonBenefits=False):
        self.check_file_path(filePath)
        self.check_mutually_exclusive_inputs(mergeLinkBuses)
        self._dispatchMode = dispatchMode
        self._optimizationType = opt
        self._mergeBuses = mergeBuses
        self.__noOfBuildings = numberOfBuildings
        logging.info("Defining the energy network from the excel file: {}".format(filePath))
        data = pd.ExcelFile(filePath)
        initial_nodal_data = self.get_nodal_data_from_Excel(data)
        data.close()
        other_data_reader = read.ProfileAndOtherDataReader()

        nodesData = other_data_reader.read_profiles_and_other_data(initial_nodal_data, filePath, numberOfBuildings, clusterSize,
                                                                   self.timeindex)
        self.set_using_nodal_data(clusterSize, filePath, includeCarbonBenefits, nodesData, mergeBuses,
                                  mergeHeatSourceSink, mergeLinkBuses, numberOfBuildings, opt)

    def set_using_nodal_data(self, clusterSize, filePath, includeCarbonBenefits, nodesData: dict[str, pd.DataFrame], mergeBuses,
                             mergeHeatSourceSink, mergeLinkBuses, numberOfBuildings, opt,
                             grouped_network: bool = False
                             ):
        if grouped_network and not isinstance(self, EnergyNetworkGroup):
            logging.error(f"'grouped_network' flag should not be used without using {EnergyNetworkGroup.__name__}.")

        self._convertNodes(nodesData, opt, mergeLinkBuses, mergeBuses, mergeHeatSourceSink, includeCarbonBenefits,
                           clusterSize)
        logging.info("Nodes from {} successfully converted".format(filePath))

        if grouped_network:
            assert isinstance(self, EnergyNetworkGroup)
            self._addLinks(nodesData["links"], numberOfBuildings, mergeLinkBuses)
            logging.info("Links successfully added to the energy network")

        self.add(*self._nodesList)
        logging.info("Nodes successfully added to the energy network")

    def check_mutually_exclusive_inputs(self, mergeLinkBuses):
        if mergeLinkBuses and self._temperatureLevels:
            logging.error("The options mergeLinkBuses and temperatureLevels should not be set True at the same time. "
                          "This use case is not supported in the present version of optihood. Only one of "
                          "mergeLinkBuses and temperatureLevels should be set to True.")

    @staticmethod
    def check_file_path(filePath):
        if not filePath or not os.path.isfile(filePath):
            logging.error(f"File {filePath} not found.")

    def get_nodal_data_from_df(self, data: pd.DataFrame):
        nodes_data = self.get_nodal_data(get_data_from_df, data)

        return nodes_data

    def get_nodal_data_from_dict_with_dfs(self, data: dict):
        """ Facade as the interface is the same as for the DataFrame version. """
        nodes_data = self.get_nodal_data(get_data_from_df, data)

        return nodes_data

    def get_nodal_data_from_Excel(self, data: pd.ExcelFile):
        nodes_data = self.get_nodal_data(get_data_from_excel_file, data)

        return nodes_data

    @staticmethod
    def get_nodal_data(func, data: _tp.Union[pd.ExcelFile, pd.DataFrame, dict]):
        f""" Uses a specified function to get the nodal data from a specified entry point.
            e.g.:
            func = {get_data_from_excel_file}
            func = {get_data_from_df}            
        """
        nodes_data = {
            _ent.NodeKeys.buses.value: func(data, _ent.NodeKeys.buses.value),
            _ent.NodeKeys.grid_connection.value: func(data, _ent.NodeKeys.grid_connection.value),
            _ent.NodeKeys.commodity_sources.value: func(data, _ent.NodeKeys.commodity_sources.value),
            _ent.NodeKeys.solar.value: func(data, _ent.NodeKeys.solar.value),
            _ent.NodeKeys.transformers.value: func(data, _ent.NodeKeys.transformers.value),
            _ent.NodeKeys.demand.value: func(data, _ent.NodeKeys.demand.value),
            _ent.NodeKeys.storages.value: func(data, _ent.NodeKeys.storages.value),
            _ent.NodeKeys.stratified_storage.value: func(data, _ent.NodeKeys.stratified_storage.value),
            _ent.NodeKeys.profiles.value: func(data, _ent.NodeKeys.profiles.value),
        }

        try:
            # This sheet is usually not included for single buildings.
            nodes_data[_ent.NodeKeys.links.value] = func(data, _ent.NodeKeys.links.value)
        except:  # multiple exceptions possible, each of which leads to the same conclusion.
            logging.warning("No data found for links. Moving on without including links.")
        try:
            if "ice_storage" in data.sheet_names:
                nodes_data.update({_ent.NodeKeys.ice_storage.value: func(data, _ent.NodeKeys.ice_storage.value)})
        except:  # multiple exceptions possible, each of which leads to the same conclusion.
            logging.warning("No data found for an ice storage. Moving on without including one.")

        return nodes_data

    def _checkStorageTemperatureGradient(self, temp_c):
        listTemperatures = [temp_c]
        listTemperatures.extend(self.__operationTemperatures)
        gradient = [h - l for l, h in zip(listTemperatures, listTemperatures[1:])]
        return all(x == gradient[0] for x in gradient)

    def _convertNodes(self, data, opt, mergeLinkBuses, mergeBuses, mergeHeatSourceSink, includeCarbonBenefits, clusterSize):
        if not data:
            logging.error("Nodes data is missing.")
        self.__temperatureAmb = np.array(data["weather_data"]["tre200h0"])
        self.__temperatureGround = np.array(data["weather_data"]["ground_temp"])
        if self._temperatureLevels:
            try:
                self.__operationTemperatures = [float(t) for t in data["stratified_storage"].loc["thermalStorage", "temp_h"].split(",")] # temperatures should be saved from low to high separated by commas in the excel input file
                if not self._checkStorageTemperatureGradient(data["stratified_storage"].loc["thermalStorage", "temp_c"]):
                    raise ValueError('Temperature Gradient in each layer of the thermal storage is not equal')
            except AttributeError:
                logging.error("The column temp_h of stratified_storage excel sheet should contain comma separated temperatures"
                              "when temperatureLevels is set to True. The temperatures should be ordered starting from low"
                              "to high and the label of storage should be thermalStorage.")
            self.__temperatureSH = self.__operationTemperatures[0]
            self.__temperatureDHW = self.__operationTemperatures[-1]
        else:
            self.__temperatureSH = data["stratified_storage"].loc["shStorage", "temp_h"]
            self.__temperatureDHW = data["stratified_storage"].loc["dhwStorage", "temp_h"]
            self.__operationTemperatures = [self.__temperatureSH, self.__temperatureDHW]
        self._coolingTemperatures = []
        if "coolingBufferStorage" in data["stratified_storage"].index:
            self._temperatureHigh = data["stratified_storage"].loc["coolingBufferStorage", "temp_h"]
            self._temperatureLow = data["stratified_storage"].loc["coolingBufferStorage", "temp_c"]
            self._coolingTemperatures = [self._temperatureLow, self._temperatureHigh]

        # Transformers conversion factors input power - output power
        for comp in _ent.TransformerTypes:
            mask = data[_ent.NodeKeys.transformers.value][_ent.TransformerLabels.label.value].str.startswith(comp)
            df = data[_ent.NodeKeys.transformers.value][mask]  # Filtered DataFrame
            if not df.empty:
                eff = df[_ent.TransformerLabels.efficiency.value].iloc[0]
                if comp == _ent.TransformerTypes.CHP.value:
                    # electricity bus comes before heating bus(es) for CHP
                    value = float(eff.split(",")[1])
                elif isinstance(eff, str) and comp in [_ent.TransformerTypes.GasBoiler.value, _ent.TransformerTypes.OilBoiler.value, _ent.TransformerTypes.BiomassBoiler.value]:
                    value = float(eff.split(",")[0])
                else:
                    value = float(eff)
                if comp == _ent.TransformerTypes.ElectricRod.value:
                    if df[_ent.TransformerLabels.active.value].iloc[0] == 1:
                        # TODO: check why ElectricRod is not assumed active like other components
                        self.__technology_efficiency[comp] = value
                else:
                    self.__technology_efficiency[comp] = value

        # Storage conversion L - kWh to display the L value
        if self._temperatureLevels:
            self.__Ltank = self._rho * self._c * (self.__operationTemperatures[1] - self.__operationTemperatures[0]) / 3600
        else:
            self.__LgenericStorage = {}
            self.__Lsh = self._rho * self._c * (self.__temperatureSH - data["stratified_storage"].loc["shStorage", "temp_c"]) / 3600
            self.__Ldhw = self._rho * self._c * (self.__temperatureDHW - data["stratified_storage"].loc["dhwStorage", "temp_c"]) / 3600
            if data['storages']['label'].str.match(r'^tankGenericStorage(\d+)?$').any():
                self.__LgenericStorage['tankStorage'] = data["stratified_storage"].loc["tankGenericStorage", "energy_density_per_m3"]/1000
            if data['storages']['label'].str.match(r'^pitGenericStorage(\d+)?$').any():
                self.__LgenericStorage['pitStorage'] = data["stratified_storage"].loc["pitGenericStorage", "energy_density_per_m3"]/1000
            if data['storages']['label'].str.match(r'^boreholeGenericStorage(\d+)?$').any():
                self.__LgenericStorage['boreholeStorage'] = data["stratified_storage"].loc["boreholeGenericStorage", "energy_density_per_m3"]/1000
            if data['storages']['label'].str.match(r'^aquifierGenericStorage(\d+)?$').any():
                self.__LgenericStorage['aquifierStorage'] = data["stratified_storage"].loc["aquifierGenericStorage", "energy_density_per_m3"]/1000
        # For the pit storage (non-generic model which includes losses)
        if data['storages']['label'].str.match(r'^pitStorage(\d+)?$').any():
            self.__LgenericStorage['pitStorageModel'] = self._rho * self._c * (data["stratified_storage"].loc["pitStorage", "temp_h"]
                                                 - data["stratified_storage"].loc["pitStorage", "temp_c"]) / 3600
        # For the tank storage (non-generic model which includes losses)
        if data['storages']['label'].str.match(r'^tankStorage(\d+)?$').any():
            self.__LgenericStorage['tankStorageModel'] = self._rho * self._c * (data["stratified_storage"].loc["tankStorage", "temp_h"]
                                                    - data["stratified_storage"].loc["tankStorage", "temp_c"]) / 3600
        # Storage conversion m3 - kWh for ice storage
        if data['storages']['label'].str.match(r'^iceStorage(\d+)?$').any():
            self.__m3IceStorage = self._rho * 1000 * self._c * (10 - 0) / 3600
        self._generate_transformer_sh_output_flow_dict(data["transformers"])
        self._addBuildings(data, opt, mergeLinkBuses, mergeBuses, mergeHeatSourceSink, includeCarbonBenefits, clusterSize)

    def _generate_transformer_sh_output_flow_dict(self, energy_conversion_tech_data):
        """
        Creates a dictionary that maps each transformer component (transformer label and
        associated building number as dict keys) to its corresponding space heating output
        bus label.

        The mapping is stored in `self._transformer_sh_output_flow`

        This mapping is needed to update the obtained input capacities in the results to output capacities
        """
        self._transformer_sh_output_flow = {}
        for i, r in energy_conversion_tech_data.iterrows():
            dict_key = f"{r[_ent.TransformerLabels.label]}__Building{r[_ent.TransformerLabels.building]}"
            if r[_ent.TransformerLabels.label].startswith(_ent.TransformerTypes.CHP):
                sh_output_bus_label = f"{r[_ent.TransformerLabels.to].split(",")[1]}__Building{r[_ent.TransformerLabels.building]}"
            else:
                sh_output_bus_label = f"{r[_ent.TransformerLabels.to].split(",")[0]}__Building{r[_ent.TransformerLabels.building]}"
            self._transformer_sh_output_flow[dict_key] = sh_output_bus_label

    def _addBuildings(self, data, opt, mergeLinkBuses, mergeBuses, mergeHeatSourceSink, includeCarbonBenefits, clusterSize):
        numberOfBuildings = max(data["buses"]["building"])
        self.__buildings = [Building('Building' + str(i + 1)) for i in range(numberOfBuildings)]
        storageParams = {}
        for s in [_ent.NodeKeys.stratified_storage.value, _ent.NodeKeys.ice_storage.value]:
            if s in data:
                storageParams.update({s: data[s]})
        for b in self.__buildings:
            buildingLabel = b.getBuildingLabel()
            i = int(buildingLabel[8:])
            if mergeLinkBuses: b.linkBuses(busesToMerge=mergeBuses)
            if i == 1:
                busDictBuilding1 = b.addBus(data["buses"][data["buses"]["building"] == i], opt, mergeLinkBuses, mergeHeatSourceSink, data["electricity_impact"], clusterSize, includeCarbonBenefits)
            else:
                b.addBus(data["buses"][data["buses"]["building"] == i], opt, mergeLinkBuses, mergeHeatSourceSink, data["electricity_impact"], clusterSize, includeCarbonBenefits)
            if (mergeLinkBuses or mergeHeatSourceSink) and i!=1:
                if mergeLinkBuses: merge = 'links'
                if mergeHeatSourceSink: merge = 'heatSourceSink'
                b.addToBusDict(busDictBuilding1, merge)
            b.addGridSeparation(data["grid_connection"][data["grid_connection"]["building"] == i], mergeLinkBuses)
            natGasCost = natGasImpact = None
            if "natGas_cost" in data:
                natGasCost = data["natGas_cost"]
                natGasImpact = data["natGas_impact"]
            fixed_sources_data = None
            if "fixed_sources" in data:
                fixed_sources_data = {k: v for k, v in data["fixed_sources"].items() if buildingLabel in k}
            b.addSource(data["commodity_sources"][data["commodity_sources"]["building"] == i], data["electricity_impact"], data["electricity_cost"], natGasCost, natGasImpact, fixed_sources_data, opt, mergeHeatSourceSink, mergeLinkBuses)
            if i in data["building_model"]:
                bmdata = data["building_model"][i]
            else:
                bmdata = {}
            b.addSink(data["demand"][data["demand"]["building"] == i], data["demandProfiles"][i], bmdata, mergeLinkBuses, mergeHeatSourceSink, self._temperatureLevels)
            b.addTransformer(data["transformers"][data["transformers"]["building"] == i], [self.__operationTemperatures, self._coolingTemperatures], self.__temperatureAmb, self.__temperatureGround, opt, mergeLinkBuses, mergeHeatSourceSink, self._dispatchMode, self._temperatureLevels)
            storageList = b.addStorage(data["storages"][data["storages"]["building"] == i], storageParams, self.__temperatureAmb, opt, mergeLinkBuses, self._dispatchMode, self._temperatureLevels)
            b.addSolar(data["solar"][(data["solar"]["building"] == i) & (data["solar"]["label"].str.match(r"^solarCollector(\d+)?$"))], data["weather_data"], opt, mergeLinkBuses, self._dispatchMode, self._temperatureLevels)
            b.addPV(data["solar"][(data["solar"]["building"] == i) & (data["solar"]["label"].str.match(r"^pv(\d+)?$"))], data["weather_data"], opt, self._dispatchMode)
            b.addPVT(data["solar"][(data["solar"]["building"] == i) & (data["solar"]["label"].str.match(r"^pvt(\d+)?$"))], data["weather_data"], opt, mergeLinkBuses, self._dispatchMode)
            self._nodesList.extend(b.getNodesList())
            self._thermalStorageList.extend(storageList)
            self.__inputs[buildingLabel] = b.getInputs()
            self.__technologies[buildingLabel] = b.getTechnologies()
            self.__costParam.update(b.getCostParam())
            self.__envParam.update(b.getEnvParam())
            self._busDict.update(b.getBusDict())
            self.__capacitiesTransformersBuilding[buildingLabel] = {}
            self.__capacitiesStoragesBuilding[buildingLabel] = {}
            self.__envImpactInputs[buildingLabel] = {}
            self.__opex[buildingLabel] = {}
            self.__envImpactTechnologies[buildingLabel] = {}

    def printNodes(self):
        print("*********************************************************")
        print("The following objects have been created from excel sheet:")
        for n in self.nodes:
            oobj = str(type(n)).replace("<class 'oemof.solph.", "").replace("'>", "")
            print(oobj + ":", n.label)
        print("*********************************************************")

    def optimize(self, numberOfBuildings, solver, envImpactlimit=None, clusterSize={},
                 options=None,   # solver options
                 optConstraints=None, #optional constraints (implemented for the moment are "roof area"
                 mergeLinkBuses=False):

        # ======================================================
        # TODO: Because we use "__" here, we break inheritance.  # pylint: disable=fixme
        #       When the EnergyNetworkGroup adds self.__noOfBuildings,
        #       it shows up as self._EnergyNetworkGroup__noOfBuildings.
        #       When the EnergyNetworkClass adds self.__noOfBuildings,
        #       it shows up as self._EnergyNetworkClass__noOfBuildings.
        #       Therefore, when a method in the EnergyNetworkClass tries to call,
        #       self.__noOfBuildings,
        #       then, it will look for self._EnergyNetworkClass__noOfBuildings.
        #       If the self is actually an EnergyNetworkGroup, then it will not find this.
        self.__noOfBuildings = numberOfBuildings
        # ======================================================

        if options is None:
            options = {"gurobi": {"tee": True,
                                  "options": {"MIPGap": 0.01}},
                       "cbc": {"tee": True},
                       "glpk": {"tee": True}}

        self._optimizationModel = solph.Model(self)
        logging.info("Optimization model built successfully")

        # add constraint to limit the environmental impacts
        self._optimizationModel, flows, transformerFlowCapacityDict, storageCapacityDict = environmentalImpactlimit(
            self._optimizationModel, keyword1="env_per_flow", keyword2="env_per_capa", limit=envImpactlimit)

        if self._temperatureLevels:
            # add constraint on storage volume capacity
            self._optimizationModel = multiTemperatureStorageCapacityConstaint(self._optimizationModel, self._thermalStorageList, self._optimizationType)

        # optional constraints (available: 'roof area')
        if optConstraints:
            for c in optConstraints:
                if c.lower() == "roof area":
                    # requires 2 additional parameters in the scenario file, tab "solar", zenit angle, roof area
                    try:
                        self._optimizationModel = roof_area_limit(self._optimizationModel,
                                                        keyword1="space", keyword2="roof_area", nb=numberOfBuildings)
                        logging.info(f"Optional constraint {c} successfully added to the optimization model")
                    except ValueError:
                        logging.error(f"Optional constraint {c} not added to the optimization model : "
                                      f"please check if PV efficiency, roof area and zenith angle are present in input "
                                      f"file")
                        pass
                if c.lower() == 'totalpvcapacity':
                    self._optimizationModel = totalPVCapacityConstraint(self._optimizationModel, numberOfBuildings)
                    logging.info(f"Optional constraint {c} successfully added to the optimization model")
        # constraint on elRod combined with HPs:
        if not np.isnan(self.__elRodEff):
            self._optimizationModel = electricRodCapacityConstaint(self._optimizationModel, numberOfBuildings)
        # constraint on PVT capacity if PVT technology is selected
        if any("pvt" in n.label for n in self.nodes):
            self._optimizationModel = PVTElectricalThermalCapacityConstraint(self._optimizationModel, numberOfBuildings)
        # constraint on STC capacity if STC technology is selected
        if any("solarCollector" in n.label for n in self.nodes):
            self._optimizationModel = STCThermalCapacityConstraint(self._optimizationModel, numberOfBuildings)
        # constraint on storage content for clustering
        """if clusterSize:
            self._optimizationModel = dailySHStorageConstraint(self._optimizationModel)"""

        logging.info("Custom constraints successfully added to the optimization model")
        logging.info("Initiating optimization using {} solver".format(solver))

        self._optimizationModel.solve(solver=solver, solve_kwargs=options[solver])

        # obtain the value of the environmental impact (subject to the limit constraint)
        # the optimization imposes an integral limit constraint on the environmental impacts
        # total environmental impacts <= envImpactlimit
        envImpact = self._optimizationModel.totalEnvironmentalImpact()

        self._optimizationResults = solph.processing.results(self._optimizationModel, remove_last_time_point=True)
        self._metaResults = solph.processing.meta_results(self._optimizationModel)
        logging.info("Optimization successful and results collected")

        # calculate capacities invested for transformers and storages (for the entire energy network and per building)
        capacitiesTransformersNetwork, capacitiesStoragesNetwork = self._calculateInvestedCapacities(self._optimizationModel, transformerFlowCapacityDict, storageCapacityDict)

        if clusterSize:
            self._postprocessingClusters(clusterSize)

        # calculate results (CAPEX, OPEX, FeedIn Costs, environmental impacts etc...) for each building
        self._calculateResultsPerBuilding(mergeLinkBuses)

        # todo: return data class
        #       - unprocessed results sheets
        #       - processed results
        #           - total
        #           - per building
        #           - envImpact, capacitiesTransformersNetwork, capacitiesStoragesNetwork
        #           - costs, en
        #       - optimization metadata
        #       -
        # def optimize(self, return_data_class=True):
        # if return_data_class:
        #     return a

        # warning("sunsetting old")
        return envImpact, capacitiesTransformersNetwork, capacitiesStoragesNetwork


    def printbuildingModelTemperatures(self, filename):
        df = pd.DataFrame()
        df["timestamp"] = self.timeindex[0:-1]
        for i in range(self.__noOfBuildings):
            bNo = i + 1
            tIndoor = [v for k, v in self._optimizationModel.SinkRCModelBlock.tIndoor.get_values().items() if
                       k[0].label.endswith(f"Building{bNo}")]
            tDistribution = [v for k, v in self._optimizationModel.SinkRCModelBlock.tDistribution.get_values().items() if
                             k[0].label.endswith(f"Building{bNo}")]
            tWall = [v for k, v in self._optimizationModel.SinkRCModelBlock.tWall.get_values().items() if
                     k[0].label.endswith(f"Building{bNo}")]
            epsilonIndoor = [v for k, v in self._optimizationModel.SinkRCModelBlock.epsilonIndoor.get_values().items() if
                             k[0].label.endswith(f"Building{bNo}")]
            tIndoor_prev = [v for k, v in self._optimizationModel.SinkRCModelBlock.tIndoor_prev.get_values().items() if
                            k[0].label.endswith(f"Building{bNo}")]
            tDistribution_prev = [v for k, v in
                                  self._optimizationModel.SinkRCModelBlock.tDistribution_prev.get_values().items() if
                                  k[0].label.endswith(f"Building{bNo}")]
            tWall_prev = [v for k, v in self._optimizationModel.SinkRCModelBlock.tWall_prev.get_values().items() if
                          k[0].label.endswith(f"Building{bNo}")]
            df[f"tIndoor_B{bNo}"] = tIndoor
            df[f"tDistribution_B{bNo}"] = tDistribution
            df[f"tWall_B{bNo}"] = tWall
            df[f"tIndoor_prev_B{bNo}"] = tIndoor_prev
            df[f"tDistribution_prev_B{bNo}"] = tDistribution_prev
            df[f"tWall_prev_B{bNo}"] = tWall_prev
            df[f"epsilonIndoor_B{bNo}"] = epsilonIndoor
        df.to_csv(filename, sep=';', index=False)

    def exportIceStorageModelParams(self, filename):
        df = pd.DataFrame()
        df["timestamp"] = self.timeindex[0:-1]
        for i in range(self.__noOfBuildings):
            bNo = i + 1
            tStor = [v for k, v in self._optimizationModel.IceStorageBlock.tStor.get_values().items() if
                       k[0].label.endswith(f"Building{bNo}")]
            mIceStor = [v for k, v in self._optimizationModel.IceStorageBlock.mIceStor.get_values().items() if
                             k[0].label.endswith(f"Building{bNo}")]
            fIce = [v for k, v in self._optimizationModel.IceStorageBlock.fIce.get_values().items() if
                     k[0].label.endswith(f"Building{bNo}")]
            iceStatus = [v for k, v in self._optimizationModel.IceStorageBlock.iceStatus.get_values().items() if
                             k[0].label.endswith(f"Building{bNo}")]
            tStor_prev = [v for k, v in self._optimizationModel.IceStorageBlock.tStor_prev.get_values().items() if
                            k[0].label.endswith(f"Building{bNo}")]
            mIceStor_prev = [v for k, v in
                                  self._optimizationModel.IceStorageBlock.mIceStor_prev.get_values().items() if
                                  k[0].label.endswith(f"Building{bNo}")]
            if tStor:
                df[f"tStor_B{bNo}"] = tStor
                df[f"mIceStor_B{bNo}"] = mIceStor
                df[f"fIce_B{bNo}"] = fIce
                df[f"iceStatus_prev_B{bNo}"] = iceStatus
                df[f"tStor_prev_B{bNo}"] = tStor_prev
                df[f"mIceStor_prev_B{bNo}"] = mIceStor_prev
        df.to_csv(filename, sep=';', index=False)

    def saveUnprocessedResults(self, resultFile=None):
        busLabelList = []
        result = {}
        for i in self.nodes:
            if str(type(i)).replace("<class 'oemof.solph.", "").replace("'>", "") == "buses._bus.Bus":
                busLabelList.append(i.label)
        for i in busLabelList:
            if "sequences" in solph.views.node(self._optimizationResults, i):
                result[i] = pd.DataFrame.from_dict(solph.views.node(self._optimizationResults, i)["sequences"])
                if resultFile is not None:
                    with pd.ExcelWriter(resultFile) as writer:
                        result[i].to_excel(writer, sheet_name=i)
        return result

    def _updateCapacityDictInputInvestment(self, transformerFlowCapacityDict):
        for inflow, outflow in list(transformerFlowCapacityDict):
            index = (inflow, outflow)
            if any(c in str(outflow) for c in _ent.TransformerTypes):
                new_out_flow = self._transformer_sh_output_flow[str(outflow)]
                new_index = (outflow,new_out_flow)
                transformerFlowCapacityDict[new_index] = transformerFlowCapacityDict.pop(index)
            if _ent.BusTypes.electricitySourceBusPVT in str(inflow):   # remove PVT electrical capacity
                transformerFlowCapacityDict.pop(index)
        return transformerFlowCapacityDict

    def _calculateInvestedCapacities(self, optimizationModel, transformerFlowCapacityDict, storageCapacityDict):
        capacitiesInvestedTransformers = {}
        capacitiesInvestedStorages = {}
        storageList = []

        # Transformers capacities

        for inflow, outflow in transformerFlowCapacityDict:
            index = (str(inflow), str(outflow))
            if (inflow, outflow) in optimizationModel.InvestmentFlowBlock.invest:
                capacitiesInvestedTransformers[index] = optimizationModel.InvestmentFlowBlock.invest[
                    inflow, outflow].value
            else:
                capacitiesInvestedTransformers[index] = optimizationModel.InvestNonConvexFlowBlock.invest[
                    inflow, outflow].value

        # Conversion into output capacity
        capacitiesInvestedTransformers = self._compensateInputCapacities(capacitiesInvestedTransformers)

        # update transformerFlowCapacityDict such that investment objects with input flows are accounted properly in the results calculation
        transformerFlowCapacityDict = self._updateCapacityDictInputInvestment(transformerFlowCapacityDict)

        # Update capacities
        for inflow, outflow in transformerFlowCapacityDict:
            index = (str(inflow), str(outflow))
            buildingLabel = str(inflow).split("__")[1]
            self.__capacitiesTransformersBuilding[buildingLabel].update({index: capacitiesInvestedTransformers[index]})

        # Storages capacities
        for x in storageCapacityDict:
            index = str(x)
            if x in storageList:  # useful when we want to implement two or more storage units of the same type
                if str(x).startswith(_ent.StorageTypes.pitStorage):
                    capacitiesInvestedStorages[index] = capacitiesInvestedStorages[index] + \
                                                        optimizationModel.GenericInvestmentStorageBlockPit.invest[x].value
                else:
                    capacitiesInvestedStorages[index] = capacitiesInvestedStorages[index] + \
                                                        optimizationModel.GenericInvestmentStorageBlock.invest[x].value
            else:
                if str(x).startswith(_ent.StorageTypes.pitStorage):
                    capacitiesInvestedStorages[index] = optimizationModel.GenericInvestmentStorageBlockPit.invest[x].value
                else:
                    capacitiesInvestedStorages[str(x)] = optimizationModel.GenericInvestmentStorageBlock.invest[x].value

        # Convert kWh into L
        capacitiesInvestedStorages = self._compensateStorageCapacities(capacitiesInvestedStorages)

        removeKeysList = []
        if self._temperatureLevels:
            if any("thermalStorage" in key for key in capacitiesInvestedStorages):
                for b in range(len(self.__buildings)):
                    buildingLabel = "Building" + str(b + 1)
                    invest = 0
                    for layer in self.__operationTemperatures:
                        if f"thermalStorage{int(layer)}__" + buildingLabel in capacitiesInvestedStorages:
                            invest += capacitiesInvestedStorages[f"thermalStorage{int(layer)}__" + buildingLabel]
                            removeKeysList.append(f"thermalStorage{int(layer)}__" + buildingLabel)
                    if invest > 0:
                        capacitiesInvestedStorages["thermalStorage__" + buildingLabel] = invest
                        storageCapacityDict["thermalStorage__" + buildingLabel] = invest
                        for k in removeKeysList:
                            capacitiesInvestedStorages.pop(k, None)

        # Update capacities
        for x in storageCapacityDict:
            index = str(x)
            if index in removeKeysList:
                continue
            buildingLabel = index.split("__")[1]
            if x in storageList:  # useful when we want to implement two or more storage units of the same type
                self.__capacitiesStoragesBuilding[buildingLabel].update(
                    {index: capacitiesInvestedStorages[index]})
            else:
                self.__capacitiesStoragesBuilding[buildingLabel].update({index: capacitiesInvestedStorages[index]})
            storageList.append(x)

        return capacitiesInvestedTransformers, capacitiesInvestedStorages

    def _compensateInputCapacities(self, capacitiesTransformers):
        # Input capacity -> output capacity
        for inflow_bus, transformer in list(capacitiesTransformers):
            for c in _ent.TransformerTypes:
                if transformer.startswith(c):
                    for index, value in enumerate(self.nodes):
                        if transformer == value.label:
                            if transformer.startswith(_ent.TransformerTypes.Chiller):
                                all_buses_connected_to_transformer = self.nodes[index].outputs
                            else:
                                all_buses_connected_to_transformer = self.nodes[index].conversion_factors
                            for b in all_buses_connected_to_transformer.keys():
                                if self._transformer_sh_output_flow[transformer] in b.label:
                                    capacitiesTransformers[(transformer,b.label)]= capacitiesTransformers[(inflow_bus,transformer)]*self.__technology_efficiency[c]
                                    del capacitiesTransformers[(inflow_bus,transformer)]
        return capacitiesTransformers

    def _compensateStorageCapacities(self, capacitiesStorages):
        # kWh -> L
        for storage in capacitiesStorages.keys():
            if "sh" in storage:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__Lsh
            elif "dhw" in storage:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__Ldhw
            elif "thermal" in storage and self._temperatureLevels:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__Ltank
            elif "tankGenericStorage" in storage and not self._temperatureLevels:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__LgenericStorage['tankStorage']
            elif "pitGenericStorage" in storage and not self._temperatureLevels:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__LgenericStorage['pitStorage']
            elif "boreholeGenericStorage" in storage and not self._temperatureLevels:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__LgenericStorage['boreholeStorage']
            elif "aquifierGenericStorage" in storage and not self._temperatureLevels:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__LgenericStorage['aquifierStorage']
            elif "pitStorage" in storage and not self._temperatureLevels:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__LgenericStorage['pitStorageModel']       # non-generic storage model with loss terms depending on geometry
            elif "tankStorage" in storage and not self._temperatureLevels:
                capacitiesStorages[storage] = capacitiesStorages[storage] / self.__LgenericStorage['tankStorageModel']      # non-generic storage model with loss terms depending on geometry
            elif "iceStorage" in storage:
                capacitiesStorages[storage] = capacitiesStorages[storage]/self.__m3IceStorage
        return capacitiesStorages

    def _postprocessingClusters(self, clusterSize):
        flows = [x for x in self._optimizationResults.keys() if x[1] is not None]
        for flow in flows:
            extrapolated_results = None
            for day in clusterSize:
                temp = pd.concat([self._optimizationResults[flow]['sequences'].loc[self._clusterDate[day]]] * clusterSize[day])
                if extrapolated_results is not None:
                    extrapolated_results = pd.concat([extrapolated_results, temp])
                else:
                    extrapolated_results = temp
            extrapolated_results.index = self._timeIndexReal
            extrapolated_results.columns = ['flow']
            self._optimizationResults[flow]['sequences'] = extrapolated_results

    def _calculateResultsPerBuilding(self, mergeLinkBuses):
        for b in self.__buildings:
            buildingLabel = b.getBuildingLabel()
            capacityTransformers = self.__capacitiesTransformersBuilding[buildingLabel]
            capacityStorages = self.__capacitiesStoragesBuilding[buildingLabel]
            technologies = self.__technologies[buildingLabel]
            inputs = self.__inputs[buildingLabel]

            # CAPital investment EXpenditure
            self.__capex[buildingLabel] = sum((self.__costParam[i][1] * (capacityTransformers[(i, o)] > 1))             # base investment cost if the technology is implemented
                                              + (capacityTransformers[(i, o)] * self.__costParam[i][0])                 # investment cost per unit capacity
                                              for i, o in capacityTransformers) + \
                                          sum((self.__costParam[x][1] * (capacityStorages[x] > 1))                      # base investment cost if the technology is implemented
                                              + (capacityStorages[x] * self.__costParam[x][0])                          # investment cost per unit capacity
                                              for x in capacityStorages)

            electricitySourceLabel = "electricityResource" + '__' + buildingLabel
            gridBusLabel = "gridBus" + '__' + buildingLabel
            naturalGasSourceLabel = "naturalGasResource" + '__' + buildingLabel
            naturalGasBusLabel = "naturalGasBus" + '__' + buildingLabel
            if mergeLinkBuses and ("electricity" in self._mergeBuses or "electricityBus" in self._mergeBuses):
                electricityBusLabel = "electricityBus"
                excessElectricityBusLabel = "excesselectricityBus"
            else:
                electricityBusLabel = "electricityBus" + '__' + buildingLabel
                excessElectricityBusLabel = "excesselectricityBus" + '__' + buildingLabel

            if electricitySourceLabel in self.__costParam:
                costParamGridElectricity = self.__costParam[electricitySourceLabel].copy()
                costParamGridElectricity.reset_index(inplace=True, drop=True)
            else:
                costParamGridElectricity = 0

            if naturalGasSourceLabel in self.__costParam:
                costParamNaturalGas = self.__costParam[naturalGasSourceLabel].copy()
                costParamNaturalGas.reset_index(inplace=True, drop=True)
            else:
                costParamNaturalGas = 0

            if "sequences" in solph.views.node(self._optimizationResults, gridBusLabel):
                if ((electricitySourceLabel, gridBusLabel), "flow") in solph.views.node(self._optimizationResults, gridBusLabel)["sequences"]:
                    gridElectricityFlow = solph.views.node(self._optimizationResults, gridBusLabel)["sequences"][
                        (electricitySourceLabel, gridBusLabel), "flow"]
                    gridElectricityFlow.reset_index(inplace=True, drop=True)
                else:
                    gridElectricityFlow = 0
            else:
                gridElectricityFlow = 0

            if "sequences" in solph.views.node(self._optimizationResults, naturalGasBusLabel):
                if ((naturalGasSourceLabel, naturalGasBusLabel), "flow") in solph.views.node(self._optimizationResults, naturalGasBusLabel)["sequences"]:
                    naturalGasFlow = solph.views.node(self._optimizationResults, naturalGasBusLabel)["sequences"][
                        (naturalGasSourceLabel, naturalGasBusLabel), "flow"]
                    naturalGasFlow.reset_index(inplace=True, drop=True)
                else:
                    naturalGasFlow = 0
            else:
                naturalGasFlow = 0

            # OPeration EXpenditure
            self.__opex[buildingLabel].update({i[0]: sum(
                solph.views.node(self._optimizationResults, i[1])["sequences"][(i[0], i[1]), "flow"] * self.__costParam[
                    i[0]]) for i in inputs})
            c = costParamGridElectricity * gridElectricityFlow
            if isinstance(c, (int, float)):
                self.__opex[buildingLabel].update({electricitySourceLabel:c})
            else:
                self.__opex[buildingLabel].update({electricitySourceLabel: c.sum()})  # cost of grid electricity is added separately based on cost data

            c = costParamNaturalGas * naturalGasFlow
            if isinstance(c, (int, float)):
                self.__opex[buildingLabel].update({naturalGasSourceLabel: c})
            else:
                self.__opex[buildingLabel].update({naturalGasSourceLabel: c.sum()})  # cost of natural gas is added separately based on cost data

            # Feed-in electricity cost (value will be in negative to signify monetary gain...)
            if ((mergeLinkBuses and buildingLabel=='Building1') or not mergeLinkBuses) and \
                    (((electricityBusLabel, excessElectricityBusLabel), "flow") in solph.views.node(self._optimizationResults, electricityBusLabel)["sequences"]):
                self.__feedIn[buildingLabel] = sum(solph.views.node(self._optimizationResults, electricityBusLabel)
                                                   ["sequences"][(electricityBusLabel, excessElectricityBusLabel), "flow"]) * self.__costParam[excessElectricityBusLabel]
            else: # in case of merged links feed in for all buildings except Building1 is set to 0 (to avoid repetition)
                self.__feedIn[buildingLabel] = 0
            if mergeLinkBuses and ("electricity" in self._mergeBuses or "electricityInBus" in self._mergeBuses):
                elInBusLabel = 'electricityInBus'
            else:
                elInBusLabel = 'electricityInBus__'+buildingLabel
            if self._temperatureLevels:
                shOutputLabel = "heatStorageBus0__"
                dhwOutputLabel = f"heatStorageBus{len(self.__operationTemperatures)-1}__"
            else:
                shOutputLabel = "shSourceBus__"
                dhwOutputLabel = "dhwStorageBus__"

            # HP flows
            if ("HP__" + buildingLabel, shOutputLabel + buildingLabel) in capacityTransformers:
                if capacityTransformers[("HP__" + buildingLabel, shOutputLabel + buildingLabel)] > 1e-3:
                    self.__elHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, elInBusLabel)["sequences"][
                            (elInBusLabel, 'HP__'+buildingLabel), 'flow'])
                    self.__shHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, 'HP__' + buildingLabel)["sequences"][
                            ('HP__' + buildingLabel, shOutputLabel + buildingLabel), 'flow'])
                    for i in range(len(self.__operationTemperatures)-2):
                        self.__intermediateOpTempsHP[buildingLabel] = sum(
                            solph.views.node(self._optimizationResults, 'HP__' + buildingLabel)["sequences"][
                                ('HP__' + buildingLabel, f"heatStorageBus{i+1}__{buildingLabel}"), 'flow'])
                    try:
                        self.__dhwHP[buildingLabel] = sum(
                            solph.views.node(self._optimizationResults, 'HP__' + buildingLabel)["sequences"][
                                ('HP__' + buildingLabel, dhwOutputLabel + buildingLabel), 'flow'])
                    except KeyError as e:
                        """If the dhwStorage is never used, then this will not be available."""
                        pass

                    if buildingLabel in self.__dhwHP.keys():
                        self.__annualCopHP[buildingLabel] = (self.__shHP[buildingLabel] + self.__dhwHP[buildingLabel]) / (
                            self.__elHP[buildingLabel] + 1e-6)
                    else:
                        self.__annualCopHP[buildingLabel] = self.__shHP[buildingLabel] / (
                            self.__elHP[buildingLabel] + 1e-6)
                else:
                    self.__annualCopHP[buildingLabel] = 0

            # GWHP flows
            if ("GWHP__" + buildingLabel, shOutputLabel + buildingLabel) in capacityTransformers:
                if capacityTransformers[("GWHP__" + buildingLabel, shOutputLabel + buildingLabel)] > 1e-3:
                    self.__elGWHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, elInBusLabel)["sequences"][
                            (elInBusLabel, 'GWHP__' + buildingLabel), 'flow'])
                    self.__shGWHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, 'GWHP__' + buildingLabel)["sequences"][
                            ('GWHP__' + buildingLabel, shOutputLabel + buildingLabel), 'flow'])
                    for i in range(len(self.__operationTemperatures)-2):
                        self.__intermediateOpTempsGWHP[buildingLabel] = sum(
                            solph.views.node(self._optimizationResults, 'GWHP__' + buildingLabel)["sequences"][
                                ('GWHP__' + buildingLabel, f"heatStorageBus{i+1}__{buildingLabel}"), 'flow'])
                    self.__dhwGWHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, 'GWHP__' + buildingLabel)["sequences"][
                            ('GWHP__' + buildingLabel, dhwOutputLabel + buildingLabel), 'flow'])
                    self.__annualCopGWHP[buildingLabel] = (self.__shGWHP[buildingLabel] + self.__dhwGWHP[buildingLabel]) / (
                            self.__elGWHP[buildingLabel] + 1e-6)
                else:
                    self.__annualCopGWHP[buildingLabel] = 0
            else:       # splitted GSHP ( at the moment this only works at 2 temperature levels)
                self.__annualCopGWHP[buildingLabel] = []
                if (f"GWHP{str(self.__temperatureSH)}__" + buildingLabel, shOutputLabel + buildingLabel) in capacityTransformers:
                    self.__elGWHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, elInBusLabel)["sequences"][
                            (elInBusLabel, f"GWHP{str(self.__temperatureSH)}__" + buildingLabel), 'flow'])
                    self.__shGWHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, f"GWHP{str(self.__temperatureSH)}__" + buildingLabel)["sequences"][
                            (f"GWHP{str(self.__temperatureSH)}__" + buildingLabel, shOutputLabel + buildingLabel), 'flow'])
                    self.__annualCopGWHP[buildingLabel].append((self.__shGWHP[buildingLabel]) / (self.__elGWHP[buildingLabel] + 1e-6))
                if (f"GWHP{str(self.__temperatureDHW)}__" + buildingLabel, dhwOutputLabel + buildingLabel) in capacityTransformers:
                    self.__elGWHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, elInBusLabel)["sequences"][
                            (elInBusLabel, f"GWHP{str(self.__temperatureDHW)}__" + buildingLabel), 'flow'])
                    self.__dhwGWHP[buildingLabel] = sum(
                        solph.views.node(self._optimizationResults, f"GWHP{str(self.__temperatureDHW)}__" + buildingLabel)["sequences"][
                            (f"GWHP{str(self.__temperatureDHW)}__" + buildingLabel, dhwOutputLabel + buildingLabel), 'flow'])
                    self.__annualCopGWHP[buildingLabel].append((self.__dhwGWHP[
                        buildingLabel]) / (self.__elGWHP[buildingLabel] + 1e-6))

            if electricitySourceLabel in self.__envParam:
                envParamGridElectricity = self.__envParam[electricitySourceLabel].copy()
                envParamGridElectricity.reset_index(inplace=True, drop=True)
            else:
                envParamGridElectricity = 0

            if naturalGasSourceLabel in self.__envParam:
                envParamNaturalGas = self.__envParam[naturalGasSourceLabel].copy()
                envParamNaturalGas.reset_index(inplace=True, drop=True)
            else:
                envParamNaturalGas = 0

            # Environmental impact due to inputs (natural gas, electricity, etc...)
            self.__envImpactInputs[buildingLabel].update({i[0]: sum(solph.views.node(self._optimizationResults, i[1])["sequences"][(i[0], i[1]), "flow"] * self.__envParam[i[0]]) for i in inputs})
            c = envParamGridElectricity * gridElectricityFlow
            if isinstance(c, (int, float)):
                self.__envImpactInputs[buildingLabel].update({electricitySourceLabel: c})
            else:
                self.__envImpactInputs[buildingLabel].update({electricitySourceLabel: c.sum()})  # impact of grid electricity is added separately based on LCA data

            c = envParamNaturalGas * naturalGasFlow
            if isinstance(c, (int, float)):
                self.__envImpactInputs[buildingLabel].update({naturalGasSourceLabel: c})
            else:
                self.__envImpactInputs[buildingLabel].update({naturalGasSourceLabel: c.sum()})  # impact of natural gas is added separately

            # Environmental impact due to technologies (converters, storages)
            # calculated by adding both environmental impact per capacity and per flow (electrical flow or heat flow)
            self.__envImpactTechnologies[buildingLabel].update({i: capacityTransformers[(i, o)] * self.__envParam[i][2] +
                                                                   sum(sum(solph.views.node(self._optimizationResults,
                                                                                            t[0])["sequences"][((t[1], t[0]), "flow")] *
                                                                           self.__envParam[t[1]][1] * ('electricityBus' in t[0]) + \
                                                                           solph.views.node(self._optimizationResults,
                                                                                            t[0])["sequences"][((t[1], t[0]), "flow")] *
                                                                           self.__envParam[t[1]][0] * ('electricityBus' not in t[0]))
                                                                       for t in technologies if t[1] == i)
                                                                for i, o in capacityTransformers})
            self.__envImpactTechnologies[buildingLabel].update({x: capacityStorages[x] * self.__envParam[x][2] +
                                                                   sum(sum(
                                                                       solph.views.node(self._optimizationResults,
                                                                                        t[0])["sequences"][((t[1], t[0]), "flow")] *
                                                                       self.__envParam[t[1]][1] * ('electricityBus' in t[0]) + \
                                                                       solph.views.node(self._optimizationResults,
                                                                                        t[0])["sequences"][((t[1], t[0]), "flow")] *
                                                                       self.__envParam[t[1]][0] * ('electricityBus' not in t[0]))
                                                                       for t in technologies if t[1] == x)
                                                                for x in capacityStorages if "thermalStorage" not in x})
            if self._temperatureLevels:
                impactThermalStorage = 0
                for x in capacityStorages:
                    if "thermalStorage" in x:
                        impactThermalStorage += capacityStorages[x] * self.__envParam["thermalStorage__"+x.split("__")[1]][2] + \
                                                                   sum(sum(
                                                                       solph.views.node(self._optimizationResults,
                                                                                        t[0])["sequences"][((t[1], t[0]), "flow")] *
                                                                       self.__envParam["thermalStorage__"+x.split("__")[1]][1] * ('electricityBus' in t[0]) + \
                                                                       solph.views.node(self._optimizationResults,
                                                                                        t[0])["sequences"][((t[1], t[0]), "flow")] *
                                                                       self.__envParam["thermalStorage__"+x.split("__")[1]][0] * ('electricityBus' not in t[0]))
                                                                       for t in technologies if (x.split("__")[0] in t[1].split("__")[0]
                                                                                                 and x.split("__")[1] == t[1].split("__")[1]))
                        self.__envImpactTechnologies[buildingLabel].update({x: impactThermalStorage})

    def printMetaresults(self):
        self._communicate_meta_results(self._metaResults, pp.pprint)
        return self._metaResults

    def log_meta_results(self) -> None:
        self._communicate_meta_results(self._metaResults, logging.info)

    @staticmethod
    def _communicate_meta_results(meta_results, communication_method: callable) -> None:
        communication_method("")
        communication_method("Meta Results:")
        communication_method(meta_results)

    def calcStateofCharge(self, type, building):
        if "thermalStorage" in type:
            for i,temperature in enumerate(self.__operationTemperatures):
                thermal_type = type + str(temperature)[:2]
                if thermal_type + '__' + building in self.groups:
                    storage = self.groups[thermal_type + '__' + building]
                    self._storageContent.setdefault(building, {})[thermal_type] = self._optimizationResults[(storage, None)]["sequences"]
                if temperature == self.__operationTemperatures[-1]:
                    lists = np.array(list(self._storageContent[building].values()))
                    # Calculate the sum of corresponding elements in the arrays
                    sums = np.sum(lists, axis=0)
                    self._storageContent[building][thermal_type][f'Overall_storage_content_{building}'] = sums
        elif type + '__' + building in self.groups:
            storage = self.groups[type + '__' + building]
            if building not in self._storageContent:
                self._storageContent[building] = self._optimizationResults[(storage, None)]["sequences"]
            else:
                self._storageContent[building] = pd.concat(
                    [self._storageContent[building], self._optimizationResults[(storage, None)]["sequences"]],
                    axis=1)
            building_number = int(building.replace("Building", ""))
            new_col_name = f"{type}__B{building_number:03d}_storage_content"
            self._storageContent[building].rename(columns={"storage_content": new_col_name}, inplace=True)

    def printInvestedCapacities(self, capacitiesInvestedTransformers, capacitiesInvestedStorages):
        storage_types = {
            _ent.StorageTypes.tankGenericStorage.value: "Tank Generic Storage",
            _ent.StorageTypes.pitStorage.value: "Pit Storage",
            _ent.StorageTypes.pitGenericStorage.value: "Pit Generic Storage",
            _ent.StorageTypes.boreholeGenericStorage.value: "Borehole Generic Storage",
            _ent.StorageTypes.aquifierGenericStorage.value: "Aquifier Generic Storage",
            _ent.StorageTypes.electricalStorage.value: "Electrical Storage",
            _ent.StorageTypes.dhwStorage.value: "DHW Storage Tank",
            _ent.StorageTypes.shStorage.value: "SH Storage Tank",
            _ent.StorageTypes.thermalStorage.value: "Multilayer Thermal Storage Tank"
        }

        transformer_types = {
            _ent.TransformerTypes.HP.value: "Air Source Heat Pump",
            _ent.TransformerTypes.GWHP.value: "Water Source Heat Pump",
            _ent.TransformerTypes.ElectricRod.value: "Electric Rod",
            _ent.TransformerTypes.CHP.value: "CHP",
            _ent.TransformerTypes.GasBoiler.value: "Gas Boiler",
            _ent.TransformerTypes.BiomassBoiler.value: "Biomass Boiler",
            _ent.TransformerTypes.OilBoiler.value: "Oil Boiler",
        }


        for b in range(len(self.__buildings)):
            buildingLabel = "Building" + str(b + 1)
            print("************** Optimized Capacities for {} **************".format(buildingLabel))

            for key, invest in capacitiesInvestedTransformers.items():
                if invest > 0.05 and key[0].endswith("__" + buildingLabel):
                    for transformer_prefix, label_base in transformer_types.items():
                        if key[0].startswith(transformer_prefix):
                            # Extract suffix (e.g., "1" in "BiomassBoiler1__Building1")
                            middle = key[0][len(transformer_prefix):-len("__" + buildingLabel)]
                            suffix = middle if middle.isdigit() else ""

                            # Apply suffix to all transformer types
                            label = f"{label_base} {suffix}" if suffix else label_base

                            unit = "kW"
                            print(f"Invested in {invest:.1f} {unit} {label}.")
                            #TODO: fix the following LOC
                            # printing of Annual COPs does not work currently
                            # most likely due to perfect matching of the string "HP" or "GWHP" while defining self.__annualCopHP
                            if transformer_prefix==_ent.TransformerTypes.HP.value and False:
                                print("     Annual COP = {:.1f}".format(self.__annualCopHP[buildingLabel]))
                            if transformer_prefix == _ent.TransformerTypes.GWHP.value and False:
                                print("     Annual COP = {:.1f}".format(self.__annualCopGWHP[buildingLabel]))

            if ("heatSource_SHsolarCollector__" + buildingLabel, "solarConnectBusSH__" + buildingLabel) in capacitiesInvestedTransformers:
                invest = capacitiesInvestedTransformers[("heatSource_SHsolarCollector__" + buildingLabel, "solarConnectBusSH__" + buildingLabel)]
                if invest > 0.05:
                    print("Invested in {:.1f} m² SolarCollector.".format(invest))
            if ("pv__" + buildingLabel, "electricityProdBus__" + buildingLabel) in capacitiesInvestedTransformers:
                invest = capacitiesInvestedTransformers[("pv__" + buildingLabel, "electricityProdBus__" + buildingLabel)]
                if invest > 0.05:
                    print("Invested in {:.1f} kWp  PV.".format(invest))
            if ("heatSource_SHpvt__" + buildingLabel, "pvtConnectBusSH__" + buildingLabel) in capacitiesInvestedTransformers:
                invest = capacitiesInvestedTransformers[("heatSource_SHpvt__" + buildingLabel, "pvtConnectBusSH__" + buildingLabel)]
                if invest > 0.05:
                    print("Invested in {:.1f} m² PVT collector.".format(invest))

            for key, invest in capacitiesInvestedStorages.items():
                if invest > 0.05 and key.endswith("__" + buildingLabel):
                    for storage_prefix, label_base in storage_types.items():
                        if key.startswith(storage_prefix):
                            # Extract suffix (e.g., "1" in "dhwStorage1__building1")
                            middle = key[len(storage_prefix):-len("__" + buildingLabel)]
                            suffix = middle if middle.isdigit() else ""
                            # Apply suffix to all storage types
                            label = f"{label_base} {suffix}" if suffix else label_base
                            # Choose unit
                            unit = "kWh" if storage_prefix == _ent.StorageTypes.electricalStorage.value else "L"
                            print(f"Invested in {invest:.1f} {unit} {label}.")
            print("")

    def calculate_costs(self) -> tuple[float, float, float]:
        # TODO: reduce to functional programming style by passing required arguments.
        capex_network = sum(self.__capex["Building" + str(b + 1)] for b in range(len(self.__buildings)))
        feed_in_network = sum(self.__feedIn["Building" + str(b + 1)] for b in range(len(self.__buildings)))
        opex_network = sum(sum(self.__opex["Building" + str(b + 1)].values()) for b in range(len(self.__buildings)))
        return capex_network, feed_in_network, opex_network

    def printCosts(self) -> None:
        capex_network, feed_in_network, opex_network = self.calculate_costs()
        self._communicate_costs(capex_network, feed_in_network, opex_network, print)

    def log_costs(self) -> None:
        capex_network, feed_in_network, opex_network = self.calculate_costs()
        self._communicate_costs(capex_network, feed_in_network, opex_network, logging.info)

    @staticmethod
    def _communicate_costs(capex_network: float, feed_in_network: float, opex_network: float,
                           communication_method: callable) -> None:
        communication_method(f"Investment Costs for the system: {capex_network} CHF")
        communication_method(f"Feed In Costs for the system: {feed_in_network} CHF")
        communication_method(f"Operation Costs for the system: {opex_network} CHF")
        communication_method(f"Total Costs for the system: {capex_network + feed_in_network + opex_network} CHF")

    def getTotalCosts(self):
        # TODO: check whether this can be removed.
        capexNetwork, feedinNetwork, opexNetwork = self.calculate_costs()
        return capexNetwork + opexNetwork + feedinNetwork

    def printEnvImpacts(self) -> None:
        env_impact_inputs_network, env_impact_technologies_network = self.calculate_environmental_impacts()
        self._communicate_environmental_impacts(env_impact_inputs_network, env_impact_technologies_network, print)

    def calculate_environmental_impacts(self):
        # TODO: reduce to functional programming style by passing required arguments.
        env_impact_inputs_network = sum(  # operation related impacts, as in "input resources"
            sum(self.__envImpactInputs["Building" + str(b + 1)].values()) for b in range(len(self.__buildings)))
        env_impact_technologies_network = sum(
            sum(self.__envImpactTechnologies["Building" + str(b + 1)].values()) for b in range(len(self.__buildings)))
        return env_impact_inputs_network, env_impact_technologies_network

    def log_environmental_impacts(self) -> None:
        env_impact_inputs, env_impact_technologies = self.calculate_environmental_impacts()
        self._communicate_environmental_impacts(env_impact_inputs, env_impact_technologies, logging.info)

    @staticmethod
    def _communicate_environmental_impacts(env_impact_inputs_network, env_impact_technologies_network,
                                           communication_method: callable) -> None:
        communication_method(f"Environmental impact from input resources for the system: "
                             f"{env_impact_inputs_network} kg CO2 eq")
        communication_method(f"Environmental impact from energy conversion and storage technologies for the system: "
                             f"{env_impact_technologies_network} kg CO2 eq")
        communication_method(f"Total: {env_impact_inputs_network + env_impact_technologies_network} kg CO2 eq")

    def getTotalEnvImpacts(self):
        # TODO: check whether this can be removed.
        env_impact_inputs_network, env_impact_technologies_network = self.calculate_environmental_impacts()
        return env_impact_technologies_network + env_impact_inputs_network

    def exportToExcel(self, file_name, mergeLinkBuses=False):
        hSB_sheet = [] #Special sheet for the merged heatStorageBus
        # Create a mapping between storage content names and their corresponding storage types
        storage_mapping = {
            "SH": "shStorage",
            "DHW": "dhwStorage",
            "Pit": "pitStorage",
            "Borehole": "boreholeStorage",
            "Aquifer": "aquiferStorage",
            "Tank": "tankStorage",
            "Battery": "electricalStorage"
        }

        for i in range(1, self.__noOfBuildings + 1):
            building_label = f"Building{i}"
            if self._temperatureLevels:
                self.calcStateofCharge("thermalStorage", building_label)
            else:
                if not hasattr(self, "_storage_content"):
                    self._storage_content = {}
                for storage_type in storage_mapping.values():
                    for group_key in self.groups:
                        if not isinstance(group_key, str):
                            continue
                        # Check: starts with storage_type, ends with __BuildingX, and only digits (or nothing) in between
                        if group_key.endswith(f"__{building_label}") and group_key.startswith(storage_type):
                            # Extract the part after storage_type and before __BuildingX
                            suffix = group_key[len(storage_type):group_key.index(f"__{building_label}")]
                            if suffix == "" or suffix.isdigit():
                                storage_instance = group_key.split("__")[0]
                                self.calcStateofCharge(storage_instance, building_label)

            hSB_sheet.append(f'heatStorageBus_Building{i}') #name of the different heatStorageBuses

        with pd.ExcelWriter(file_name, engine='openpyxl') as writer:
            busLabelList = []
            for i in self.nodes:
                if "buses._bus.Bus" in str(type(i)).replace("<oemof.solph.", "").replace("'>", ""):
                    busLabelList.append(i.label)
            # writing results of each bus into excel
            result_hSB = pd.DataFrame()
            hSB_building = 0
            for i in busLabelList:
                if "dummy" not in i: #don't print the dummy results anymore
                    if "domesticHotWaterBus" in i:  # special case for DHW bus (output from transformers --> dhwStorageBus --> DHW storage --> domesticHotWaterBus --> DHW Demand)
                        if not mergeLinkBuses:
                            dhwStorageBusLabel = "dhwStorageBus__" + i.split("__")[1]
                            resultDHW = pd.DataFrame.from_dict(solph.views.node(self._optimizationResults, i)["sequences"])  # result sequences of DHW bus
                            resultDHWStorage = pd.DataFrame.from_dict(solph.views.node(self._optimizationResults, dhwStorageBusLabel)["sequences"])  # result sequences of DHW storage bus
                            result = pd.concat([resultDHW, resultDHWStorage], axis=1, sort=True)
                        else:
                            result = pd.DataFrame.from_dict(solph.views.node(self._optimizationResults, i)["sequences"])  # result sequences of DHW bus
                    elif mergeLinkBuses and "dhwStorageBus" in i and "sequences" in solph.views.node(self._optimizationResults, i):
                        result = pd.DataFrame.from_dict(solph.views.node(self._optimizationResults, i)["sequences"])  # result sequences of DHW storage bus
                    elif "dhwStorageBus" not in i:  # for all the other buses except DHW storage bus (as it is already considered with DHW bus)
                        if 'sequences' in solph.views.node(self._optimizationResults, i):

                            result = pd.DataFrame.from_dict(solph.views.node(self._optimizationResults, i)["sequences"])

                            if "heatStorageBus" in i and i.split("__")[1] in self._storageContentTS:
                                # result = result[result.columns[[0, 3]]]  # get rid of 'status' and 'status_nominal' columns
                                result_hSB = pd.concat([result_hSB, result], axis=1, sort=True)
                                # for j, temperature in enumerate(self.__operationTemperatures): #this was used for the intermediate storage content (layer by layer)
                                #     if f's{j}' in i.split("__")[0]:
                                #         result = pd.concat([result, self._storageContentTS[i.split("__")[1]][
                                #             f'thermalStorage{int(temperature)}']], axis=1, sort=True)
                                if f's{len(self.__operationTemperatures)-1}' in i.split("__")[0]: #if we are at the last storage layer, then add the overall storage content of the building
                                    result_hSB = pd.concat([result_hSB, self._storageContentTS[i.split("__")[1]][
                                             f'thermalStorage{int(self.__operationTemperatures[-1])}'][f'Overall_storage_content_{i.split("__")[1]}']], axis=1, sort=True)

                        else:
                            continue
                    result[result < 0.001] = 0      # to resolve the issue of very low values in the results in certain cases, values less than 1 Watt would be replaced by 0
                    if (mergeLinkBuses or "dhwStorageBus" not in i) and "heatStorageBus" not in i:
                        result.to_excel(writer, sheet_name=i)
                    elif f"heatStorageBus{len(self.__operationTemperatures)-1}" in i: #Special case for merging the heatStorageBuses (only if at the higher thermal layer)
                        result_hSB.to_excel(writer, sheet_name= hSB_sheet[hSB_building])
                        result_hSB = pd.DataFrame() #Clean the temporary heatStorageBus sheet
                        hSB_building += 1  # Prepare for the next building

            # writing the storage content, costs and environmental impacts (of different components...) for each building
            for b in self.__buildings:
                buildingLabel = b.getBuildingLabel()

                self._storageContent[buildingLabel].to_excel(writer, sheet_name="storage_content__" + buildingLabel)

                costs = self.__opex[buildingLabel]
                costs.update({"Investment": self.__capex[buildingLabel],
                              "Feed-in": self.__feedIn[buildingLabel]})

                costsBuilding = pd.DataFrame.from_dict(costs, orient='index')
                costsBuilding.to_excel(writer, sheet_name="costs__" + buildingLabel)

                envImpact = self.__envImpactInputs[buildingLabel].copy()
                envImpact.update(self.__envImpactTechnologies[buildingLabel])

                envImpactBuilding = pd.DataFrame.from_dict(envImpact, orient='index')
                envImpactBuilding.to_excel(writer, sheet_name="env_impacts__" + buildingLabel)

                capacitiesStorages = self.__capacitiesStoragesBuilding[buildingLabel].copy()
                capacitiesStorages.update(self.__capacitiesStoragesBuilding[buildingLabel])

                capacitiesStoragesBuilding = pd.DataFrame.from_dict(capacitiesStorages, orient='index')
                capacitiesStoragesBuilding.to_excel(writer, sheet_name="capStorages__" + buildingLabel)

                capacitiesTransformers = self.__capacitiesTransformersBuilding[buildingLabel].copy()
                capacitiesTransformers.update(self.__capacitiesTransformersBuilding[buildingLabel])

                capacitiesTransformersBuilding = pd.DataFrame.from_dict(capacitiesTransformers, orient='index')
                capacitiesTransformersBuilding.to_excel(writer, sheet_name="capTransformers__" + buildingLabel)

    def set_from_csv(self, input_data_dir: _pl.Path, nr_of_buildings: int, clusterSize={}, opt: str = "costs",
                   mergeLinkBuses: bool = False, mergeBuses: _tp.Optional[_tp.Sequence[str]] = None,
                   mergeHeatSourceSink: bool = False, dispatchMode: bool = False, includeCarbonBenefits: bool = False):

        self.check_dir_path(input_data_dir)
        self.check_mutually_exclusive_inputs(mergeLinkBuses)
        self._dispatchMode = dispatchMode
        self._optimizationType = opt
        self._mergeBuses = mergeBuses
        self.__noOfBuildings = nr_of_buildings
        logging.info(f"Defining the energy network from the input files: {input_data_dir}")

        csvReader = _re.CsvScenarioReader(input_data_dir)
        initial_nodal_data = csvReader.read_scenario()

        other_data_reader = read.ProfileAndOtherDataReader()
        nodal_data = other_data_reader.read_profiles_and_other_data(initial_nodal_data, input_data_dir, nr_of_buildings,
                                                                    clusterSize, self.timeindex)
        self.set_using_nodal_data(clusterSize, input_data_dir, includeCarbonBenefits, nodal_data, mergeBuses,
                                  mergeHeatSourceSink, mergeLinkBuses, nr_of_buildings, opt)

    @staticmethod
    def check_dir_path(dir_path: _pl.Path):
        if not dir_path or not os.path.isdir(dir_path):
            logging.error(f"Directory {dir_path} not found.")


class EnergyNetworkIndiv(EnergyNetworkClass):
    # TODO: sunsetted, please use parent instead  # pylint: disable=fixme

    def __init__(self, timestamp):
        warnings.warn(f'"EnergyNetworkIndiv" will be sunsetted. Please use {__name__}.{EnergyNetworkClass.__name__} instead.')
        super().__init__(timestamp)

    @staticmethod
    def createScenarioFile(configFilePath, excelFilePath, building, numberOfBuildings):
        warnings.warn(f'"EnergyNetworkIndiv.createScenarioFile" will be sunsetted. Please use '
                      f'{_wr.__name__}.{_wr.ScenarioFileWriterExcel.__name__} or '
                      f'{_wr.__name__}.{_wr.ScenarioFileWriterCSV} instead.')

        scenarioFileWriter = _wr.ScenarioFileWriterExcel(configFilePath, building_nrs=building, version="individual")
        scenarioFileWriter.write(excelFilePath)


class EnergyNetworkGroup(EnergyNetworkClass):
    @staticmethod
    def createScenarioFile(configFilePath, excelFilePath, numberOfBuildings):
        warnings.warn(f'"EnergyNetworkGroup.createScenarioFile" will be sunsetted. Please use '
                      f'{_wr.__name__}.{_wr.ScenarioFileWriterExcel.__name__} or '
                      f'{_wr.__name__}.{_wr.ScenarioFileWriterCSV} instead.')

        scenarioFileWriter = _wr.ScenarioFileWriterExcel(configFilePath, nr_of_buildings=numberOfBuildings,
                                                         version="grouped")
        scenarioFileWriter.write(excelFilePath)

    def setFromExcel(self, filePath, numberOfBuildings, clusterSize={}, opt="costs", mergeLinkBuses=False, mergeBuses=None, mergeHeatSourceSink=False, dispatchMode=False, includeCarbonBenefits=False):
        self.check_file_path(filePath)
        logging.info("Defining the energy network from the excel file: {}".format(filePath))
        self.check_mutually_exclusive_inputs(mergeLinkBuses)
        self._dispatchMode = dispatchMode
        self._optimizationType = opt
        self._mergeBuses = mergeBuses
        self.__noOfBuildings = numberOfBuildings
        data = pd.ExcelFile(filePath)
        initial_nodal_data = self.get_nodal_data_from_Excel(data)
        data.close()

        other_data_reader = read.ProfileAndOtherDataReader()
        nodesData = other_data_reader.read_profiles_and_other_data(initial_nodal_data, filePath, numberOfBuildings, clusterSize,
                                                                   self.timeindex)

        self.set_using_nodal_data(clusterSize, filePath, includeCarbonBenefits, nodesData, mergeBuses,
                                  mergeHeatSourceSink, mergeLinkBuses, numberOfBuildings, opt, grouped_network=True)

    def set_from_csv(self, input_data_dir: _pl.Path, nr_of_buildings: int, clusterSize={}, opt: str = "costs",
                   mergeLinkBuses: bool = False, mergeBuses: _tp.Optional[_tp.Sequence[str]] = None,
                   mergeHeatSourceSink: bool = False, dispatchMode: bool = False, includeCarbonBenefits: bool = False):

        self.check_dir_path(input_data_dir)
        self.check_mutually_exclusive_inputs(mergeLinkBuses)
        self._dispatchMode = dispatchMode
        self._optimizationType = opt
        self._mergeBuses = mergeBuses
        self.__noOfBuildings = nr_of_buildings
        logging.info(f"Defining the energy network from the input files: {input_data_dir}")

        csvReader = _re.CsvScenarioReader(input_data_dir)
        initial_nodal_data = csvReader.read_scenario()

        other_data_reader = read.ProfileAndOtherDataReader()
        nodal_data = other_data_reader.read_profiles_and_other_data(initial_nodal_data, input_data_dir, nr_of_buildings,
                                                                    clusterSize, self.timeindex)
        self.set_using_nodal_data(clusterSize, input_data_dir, includeCarbonBenefits, nodal_data, mergeBuses,
                                  mergeHeatSourceSink, mergeLinkBuses, nr_of_buildings, opt, grouped_network=True)

    def _addLinks(self, data, numberOfBuildings, mergeLinkBuses):
        """connects buses A and B (denotes a bidirectional link)"""
        if mergeLinkBuses:
            return
        for i, l in data.iterrows():
                if l["investment"]:
                    investment = solph.Investment(
                        ep_costs=l["invest_cap"],
                        nonconvex=True,
                        maximum=500000,
                        offset=l["invest_base"],
                    )
                else:
                    investment = None
                busesOut = []
                busesIn = []
                # add two buses for each building link_out and link_in
                for b in range(numberOfBuildings):
                    if "sh" in l["label"]:
                        busesOut.append(self._busDict["spaceHeatingBus" + '__Building' + str(b+1)])
                        busesIn.append(self._busDict["shDemandBus" + '__Building' + str(b+1)])
                    elif "dhw" in l["label"]:
                        busesOut.append(self._busDict["domesticHotWaterBus" + '__Building' + str(b+1)])
                        busesIn.append(self._busDict["dhwDemandBus" + '__Building' + str(b+1)])
                    elif "heat" in l["label"]:
                        busesOut.append(self._busDict["heatBus" + l["label"][-1] + '__Building' + str(b + 1)])
                        busesIn.append(self._busDict["heatDemandBus" + l["label"][-1] + '__Building' + str(b + 1)])
                    elif "electricity" in l["label"]:
                        busesOut.append(self._busDict["electricityBus" + '__Building' + str(b + 1)])
                        busesIn.append(self._busDict["electricityInBus" + '__Building' + str(b + 1)])
                    elif "dhLink" in l["label"]:
                        if "districtHeatingInBus" + '__Building' + str(b + 1) in self._busDict:
                            busesOut.append(self._busDict["districtHeatingInBus" + '__Building' + str(b + 1)])
                        busesIn.append(self._busDict["districtHeatingBus" + '__Building' + str(b + 1)])

                self._nodesList.append(Link(
                    label=l["label"],
                    inputs={busA: solph.Flow() for busA in busesOut},
                    outputs={busB: solph.Flow(investment=investment) for busB in busesIn},
                    conversion_factors={busB: l["efficiency"] for busB in busesIn}
                ))



from pyomo import environ as pyo
from optihood._helpers import *
from oemof.solph._plumbing import sequence
from math import pi
from oemof.solph.constraints.flow_count_limit import limit_active_flow_count

def dailySHStorageConstraint(om):
    """
    Function to limit the SH storage capacity to 2 days
    :param om: optimization model
    :return: om: optimization model
    """
    for s in om.NODES:
        if "shStorage" in s.label:
            constr = s.label.replace("__Building","") + "_constraint_"
            for t in om.TIMESTEPS:
                if t % 48 == 0 and t!=0:
                    setattr(
                        om,
                        constr + str(t),
                        pyo.Constraint(expr=(om.GenericInvestmentStorageBlock.storage_content[s, t] == 0)),
                    )

    return om

def multiTemperatureStorageCapacityConstaint(om, storageNodes, optType):
    """Constraint on thermal storage capacity when multiple temperature levels exist"""
    storageCapacityDict = {}
    storageMaxCapacity = {}
    storageMinCapacity = {}
    #storageBaseInvestment = {}
    for x in om.GenericInvestmentStorageBlock.INVESTSTORAGES:
        if "thermalStorage" in x.label:
            if x.label.split("__")[1] not in storageCapacityDict:
                storageCapacityDict[x.label.split("__")[1]] = [x]
                storageMaxCapacity[x.label.split("__")[1]] = [s.capacityMax for s in storageNodes if x.label.split("__")[1] in s.label][0]
                storageMinCapacity[x.label.split("__")[1]] = [s.capacityMin for s in storageNodes if x.label.split("__")[1] in s.label][0]
                #storageBaseInvestment[x.label.split("__")[1]] = [s.baseInvestment for s in storageNodes if x.label.split("__")[1] in s.label][0]
            else:
                storageCapacityDict[x.label.split("__")[1]].append(x)

    for building in storageCapacityDict:
        storageCapacity =sum(om.GenericInvestmentStorageBlock.invest[x] for x in storageCapacityDict[building])
        setattr(
            om,
             "ThermalStorage_maxConstraint" + building[8:],
            pyo.Constraint(expr=(storageCapacity <= storageMaxCapacity[building])),
        )
        setattr(
            om,
            "ThermalStorage_minConstraint" + building[8:],
            pyo.Constraint(expr=(storageCapacity >= storageMinCapacity[building])),
        )
        # if storage is selected then add offset to objective function
        """if optType == "costs":
            offsetThermalStorage = storageBaseInvestment[building]*storageCapacity         #*(storageCapacity>0.001)
            om.objective += offsetThermalStorage"""

    return om

def connectInvestmentRule(om):
    """Constraint to equate the investment objects of all the output flows of a Link"""

    elLinkOutputFlows = [(i, o) for (i, o) in om.flows if (i.label == "electricityLink" or i.label == "elLink")]
    shLinkOutputFlows = [(i, o) for (i, o) in om.flows if i.label == "shLink"]
    dhwLinkOutputFlows = [(i, o) for (i, o) in om.flows if i.label == "dhwLink"]

    if elLinkOutputFlows:
        first = om.InvestmentFlowBlock.invest[next(iter(elLinkOutputFlows))]
        for (i, o) in elLinkOutputFlows:
            expr = (first == om.InvestmentFlowBlock.invest[i, o])
            setattr(
                om,
                "elLinkConstr_" + o.label,
                pyo.Constraint(expr=expr),
            )

    if shLinkOutputFlows:
        first = om.InvestmentFlowBlock.invest[next(iter(shLinkOutputFlows))]
        for (i, o) in shLinkOutputFlows:
            expr = (first == om.InvestmentFlowBlock.invest[i, o])
            setattr(
                om,
                "shLinkConstr_" + o.label,
                pyo.Constraint(expr=expr),
            )

    if dhwLinkOutputFlows:
        first = om.InvestmentFlowBlock.invest[next(iter(dhwLinkOutputFlows))]
        for (i, o) in dhwLinkOutputFlows:
            expr = (first == om.InvestmentFlowBlock.invest[i, o])
            setattr(
                om,
                "dhwLinkConstr_" + o.label,
                pyo.Constraint(expr=expr),
            )

    return om


def environmentalImpactlimit(om, keyword1, keyword2, limit):
    """
    Based on: oemof.solph.constraints.emission_limit
    Function to limit the environmental impacts during the multi-objective optimization
    :param om: model
    :param keyword1: keyword for environmental impacts per flow, placed in a solph.Flow() object
    :param keyword2: keyword for environmental impacts per capacity installed, placed in a solph.Investment() object
    :param limit: limit not to be reached
    :return:
    """
    flows = {}
    transformerFlowCapacityDictNonConvex = {}
    transformerFlowCapacityDictConvex = {}
    storageCapacityDict1 = {}
    storageCapacityDict2 = {}
    for (i, o) in om.flows:
        if hasattr(om.flows[i, o], keyword1):
            flows[(i, o)] = om.flows[i, o]
        if hasattr(om.flows[i, o].investment, keyword2):
            if om.flows[i, o].nonconvex is not None:
                transformerFlowCapacityDictNonConvex[(i, o)] = om.flows[i, o].investment
            else:
                transformerFlowCapacityDictConvex[(i, o)] = om.flows[i, o].investment

    if hasattr(om, 'GenericInvestmentStorageBlock'):
        for x in om.GenericInvestmentStorageBlock.INVESTSTORAGES:
            if hasattr(x.investment, keyword2):
                storageCapacityDict1[x] = om.GenericInvestmentStorageBlock.invest[x]

    if hasattr(om, 'GenericInvestmentStorageBlockPit'):
        for x in om.GenericInvestmentStorageBlockPit.INVESTSTORAGES:
            if hasattr(x.investment, keyword2):
                storageCapacityDict2[x] = om.GenericInvestmentStorageBlockPit.invest[x]

    envImpact = "totalEnvironmentalImpact"

    setattr(
        om,
        envImpact,
        pyo.Expression(
            expr=sum(
                # Environmental inpact of input flows
                om.flow[inflow, outflow, t]
                * om.timeincrement[t]
                * sequence(getattr(flows[inflow, outflow], keyword1))[t]
                for (inflow, outflow) in flows
                for t in om.TIMESTEPS
            )
                 # fix Environmental impact per transformer capacity from convex and nonconvex flows
                 + sum(om.InvestmentFlowBlock.invest[inflow, outflow] * getattr(
                transformerFlowCapacityDictConvex[inflow, outflow], keyword2)
                       for (inflow, outflow) in transformerFlowCapacityDictConvex)
                 + sum(om.InvestNonConvexFlowBlock.invest[inflow, outflow] * getattr(
                transformerFlowCapacityDictNonConvex[inflow, outflow], keyword2)
                       for (inflow, outflow) in transformerFlowCapacityDictNonConvex)
                 # fix Environmental impact per storage capacity
                 + sum(om.GenericInvestmentStorageBlock.invest[x] * getattr(x.investment, keyword2) for x in
                       storageCapacityDict1)
                 + sum(om.GenericInvestmentStorageBlockPit.invest[x] * getattr(x.investment, keyword2) for x in
                       storageCapacityDict2)
        ),
    )
    if limit is not None:
        setattr(
            om,
            envImpact + "_constraint",
            pyo.Constraint(expr=(getattr(om, envImpact) <= limit)),
        )

    transformerFlowCapacityDictNonConvex.update(transformerFlowCapacityDictConvex)

    storageCapacityDict = storageCapacityDict1 | storageCapacityDict2

    return om, flows, transformerFlowCapacityDictNonConvex, storageCapacityDict


def roof_area_limit(model, keyword1, keyword2, nb):
    r"""
    Based on: oemof.solph.constraints.additional_investment_flow_limit
    Constraint to limit the capacity of solar panels installed considering the roof area available
    Parameters
    ----------
    model : oemof.solph.Model
        Model to which constraints are added.
    keyword1 : attribute to consider
        Coefficient representing the area used by one unit of capacity of a solar panel
    keyword2 : attribute to consider
        Total roof area available for panels.
    nb : int
        Number of buildings in the neighbourhood

    Note
    ----
    The Investment attribute of the considered (Investment-)flows requires an
    attribute named like keyword!
    """

    for b in range(1, nb+1):
        limit = 0
        invest_flows = {}
        for (i, o) in model.flows:
            if str(b) in str(i):
                if hasattr(model.flows[i, o].investment, keyword1):
                    invest_flows[(i, o)] = model.flows[i, o].investment
                    limit = getattr(model.flows[i, o].investment, keyword2)

        limit_name = "invest_limit_" + keyword1 + "_building" + str(b)

        setattr(
            model,
            limit_name,
            pyo.Expression(
                expr=sum(
                    model.InvestmentFlowBlock.invest[inflow, outflow]
                    * getattr(invest_flows[inflow, outflow], keyword1)
                    for (inflow, outflow) in invest_flows
                )
            ),
        )

        setattr(
            model,
            limit_name + "_constraint",
            pyo.Constraint(expr=(getattr(model, limit_name) <= limit)),
        )


    return model

def electricRodCapacityConstaint(om, numBuildings):
    """constraint to set the total capacity of electric rod equal to sum of total capacity selected for HP"""
    electricRodInputFlows = [(i, o) for (i, o) in om.flows if ("ElectricRod" in o.label)]
    airHeatPumpInputFlows = [(i, o) for (i, o) in om.flows if ("HP" in o.label and "CHP" not in o.label and "GWHP" not in o.label)]
    groundHeatPumpInputFlows = [(i, o) for (i, o) in om.flows if ("GWHP" in o.label and not any([c.isdigit() for c in o.label.split("__")[0]]))]
    groundHeatPumpOutFlows = [(i, o) for (i, o) in om.flows if ("GWHP" in i.label and any([c.isdigit() for c in i.label.split("__")[0]]))]     # for splitted GSHPs

    elRodCapacityTotal = 0
    airHeatPumpCapacityTotal = 0
    groundHeatPumpCapacityTotal = 0

    for b in range(1,numBuildings+1):
        elRodCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in electricRodInputFlows if ((f'__Building{b}') in o.label)]
        airHeatPumpCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in airHeatPumpInputFlows if ((f'__Building{b}') in o.label)]
        if groundHeatPumpInputFlows:
            groundHeatPumpCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in groundHeatPumpInputFlows if ((f'__Building{b}') in o.label)]
        else:
            groundHeatPumpCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in groundHeatPumpOutFlows if ((f'__Building{b}') in i.label)]
        if elRodCapacity:
            elRodCapacity = elRodCapacity[0]
            airHeatPumpCapacity = airHeatPumpCapacity[0] if airHeatPumpCapacity else 0
            if groundHeatPumpInputFlows:
                groundHeatPumpCapacity = groundHeatPumpCapacity[0] if groundHeatPumpCapacity else 0
            else:
                groundHeatPumpCapacity = groundHeatPumpCapacity[0] + groundHeatPumpCapacity[1] if groundHeatPumpCapacity else 0
            elRodCapacityTotal = elRodCapacityTotal + elRodCapacity
            airHeatPumpCapacityTotal = airHeatPumpCapacityTotal + airHeatPumpCapacity
            groundHeatPumpCapacityTotal = groundHeatPumpCapacityTotal + groundHeatPumpCapacity

    expr = (elRodCapacityTotal <= (airHeatPumpCapacityTotal + groundHeatPumpCapacityTotal))
    setattr(
        om,
        "electricRodSizeConstr",
        pyo.Constraint(expr=expr),
    )
    return om

def totalPVCapacityConstraint(om, numBuildings):
    pvOutFlows = [(i, o) for (i, o) in om.flows if ("pv" in i.label)]
    pvCapacityTotal = 0
    for b in range(1,numBuildings+1):
        pvCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in pvOutFlows if ((f'__Building{b}') in o.label)]
        if pvCapacity:
            pvCapacity = pvCapacity[0]
            pvCapacityTotal = pvCapacityTotal + pvCapacity
    expr = (pvCapacityTotal <= 205)
    setattr(
        om,
        "PVSizeConstr",
        pyo.Constraint(expr=expr),
    )
    return om

def PVTElectricalThermalCapacityConstraint(om, numBuildings):
    pvtElOutFlows = [(i, o) for (i, o) in om.flows if ("elSource_pvt" in i.label)]
    pvtShOutFlows = [(i, o) for (i, o) in om.flows if ("heatSource_SHpvt" in i.label)]
    pvtDhwOutFlows = [(i, o) for (i, o) in om.flows if ("heatSource_DHWpvt" in i.label)]
    pvtT2OutFlows = [(i, o) for (i, o) in om.flows if ("heatSource_T2pvt" in i.label)]
    for b in range(1, numBuildings + 1):
        elCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in pvtElOutFlows if ((f'__Building{b}') in o.label)]
        shCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in pvtShOutFlows if ((f'__Building{b}') in o.label)]
        dhwCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in pvtDhwOutFlows if ((f'__Building{b}') in o.label)]
        T2Capacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in pvtT2OutFlows if ((f'__Building{b}') in o.label)]
        areaUnitCapEl = [getattr(om.flows[i, o].investment, 'space_el') for (i, o) in pvtElOutFlows if ((f'__Building{b}') in o.label)]
        areaUnitCapSh = [getattr(om.flows[i, o].investment, 'space') for (i, o) in pvtShOutFlows if ((f'__Building{b}') in o.label)]
        if elCapacity or shCapacity:
            elCapacity = elCapacity[0]
            shCapacity = shCapacity[0]
            areaUnitCapEl = areaUnitCapEl[0]
            areaUnitCapSh = areaUnitCapSh[0]
            expr1 = (elCapacity * areaUnitCapEl == shCapacity * areaUnitCapSh)
            setattr(
                om,
                "PVTSizeConstrElTh_B" + str(b),
                pyo.Constraint(expr=expr1),
            )
            if dhwCapacity:
                dhwCapacity = dhwCapacity[0]
                expr2 = (dhwCapacity == shCapacity)
                setattr(
                    om,
                    "PVTSizeConstrDhwSh_B" + str(b),
                    pyo.Constraint(expr=expr2),
                )
            if T2Capacity:
                T2Capacity = T2Capacity[0]
                expr3 = (T2Capacity == shCapacity)
                setattr(
                    om,
                    "PVTSizeConstrT2Sh_B" + str(b),
                    pyo.Constraint(expr=expr3),
                )
    return om

def STCThermalCapacityConstraint(om, numBuildings):
    stcShOutFlows = [(i, o) for (i, o) in om.flows if ("heatSource_SHsolarCollector" in i.label)]
    stcDhwOutFlows = [(i, o) for (i, o) in om.flows if ("heatSource_DHWsolarCollector" in i.label)]
    stcT2OutFlows = [(i, o) for (i, o) in om.flows if ("heatSource_T2solarCollector" in i.label)]
    for b in range(1, numBuildings + 1):
        shCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in stcShOutFlows if ((f'__Building{b}') in o.label)]
        dhwCapacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in stcDhwOutFlows if ((f'__Building{b}') in o.label)]
        T2Capacity = [om.InvestmentFlowBlock.invest[i, o] for (i, o) in stcT2OutFlows if ((f'__Building{b}') in o.label)]
        if shCapacity:
            shCapacity = shCapacity[0]
            if dhwCapacity:
                dhwCapacity = dhwCapacity[0]
                expr1 = (dhwCapacity == shCapacity)
                setattr(
                    om,
                    "STCSizeConstrDhwSh_B" + str(b),
                    pyo.Constraint(expr=expr1),
                )

            if T2Capacity:
                T2Capacity = T2Capacity[0]
                expr2 = (T2Capacity == shCapacity)
                setattr(
                    om,
                    "STCSizeConstrT2Sh_B" + str(b),
                    pyo.Constraint(expr=expr2),
                )
    return om



import warnings
import oemof

import re
from typing import Optional, Match

# Suppress FutureWarnings from oemof
warnings.filterwarnings("ignore", category=FutureWarning, module="oemof")

def pattern_at_start_followed_by_number(pattern: str, label: str) -> Optional[Match[str]]:
   return re.match(fr"^{pattern}(\d+)?$", label)

class LabelStringManipulator:
   def __init__(self, label):
      self.full_name = label
      splitted_label = label.split("__")
      self.prefix = splitted_label[0]
      self.building = splitted_label[1]

   def strip_trailing_digits_from_prefix(self):
      return self.prefix.rstrip("0123456789")



import pathlib as _pl
import unittest as _ut

import pandas as _pd
import pytest as _pt

import optihood.Visualizer.scenario_to_visualizer as stv

# TODO: export image

# TODO: test edges individually
# TODO: include default values from stratified storage ?and profiles?
# TODO: adjust diagram to have sources on the left and demands on the right.

# list of uncertainties
# - check docs
# - demand:
#   - fixed?
#   - nominal_value
#   - building model
# - grid connection
#   - efficiency
#


class TestNodalDataExample(_ut.TestCase):
    """
    This test shows a simple example outside of our problem domain.
    It helped create the architecture and is kept to keep the system flexible for unknown future objects.

    This example comes directly from the Plotly homepage.
    http://dash.plotly.com/cytoscape/events
    """

    def setUp(self):
        """ Currently fails, as this is the only case where IDs are given directly. """
        self.maxDiff = None
        self.nodalData = stv.NodalDataExample('Los Angeles', 'van', 'hou', True, 34.03,
                                              -118.25)
        self.nodalData.id = 'la'

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'la', 'label': 'Los Angeles', "lat": -118.25, "long": 34.03},
            'position': {'x': -2365.0, 'y': -680.6}
        }
        self.assertDictEqual(result, expected_dict)

    def test_read_nodal_infos(self):
        data = {'id': 'la', 'label': 'Los Angeles', "lat": -118.25, "long": 34.03}
        result = stv.NodalDataExample.read_nodal_infos(data)
        expected_string = "{'id': 'la', 'label': 'Los Angeles', 'lat': -118.25, 'long': 34.03}"
        self.assertEqual(result, expected_string)

    def test_get_edge_infos(self):
        result = self.nodalData.get_edge_infos()
        expected_dict_0 = {'data': {'source': 'van', 'target': 'la'}, 'classes': 'unknown'}
        expected_dict_1 = {'data': {'source': 'la', 'target': 'hou'}, 'classes': 'unknown'}
        self.assertDictEqual(result[0], expected_dict_0)
        self.assertDictEqual(result[1], expected_dict_1)

    def test_read_edge_infos(self):
        with _pt.raises(NotImplementedError):
            self.nodalData.read_edge_infos({'stuff': 0})


class TestCommoditySourcesConverter(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.path = _pl.Path("..\\excels\\basic_example\\electricity_impact.csv")
        self.nodalData = stv.CommoditySourcesConverter('electricityResource', None, 'gridBus',
                                                       building=1, variable_costs=0.204, CO2_impact=self.path,
                                                       active=True)
        self.nodalDataFalse = stv.CommoditySourcesConverter('electricityResource', None, 'gridBus',
                                                            building=1, variable_costs=0.204, CO2_impact=self.path,
                                                            active=False)

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'electricityResource_B001', 'label': 'electricityResource', "building": 1,
                     "variable_costs": 0.204,
                     "CO2_impact": self.path, 'color': '#1f78b4'},
            'classes': 'source',
        }
        self.assertDictEqual(result, expected_dict)

    def test_get_nodal_infos_ignored(self):
        result = self.nodalDataFalse.get_nodal_infos()
        self.assertIsNone(result)

    def test_get_edge_infos(self):
        result = self.nodalData.get_edge_infos()
        expected_dict = {'data': {'source': 'electricityResource_B001', 'target': 'gridBus'},
                         'classes': 'electricity', }
        self.assertDictEqual(result[0], expected_dict)

    def test_get_edge_infos_ignored(self):
        result = self.nodalDataFalse.get_edge_infos()
        self.assertListEqual(result, [])

    def test_set_from_dataFrame(self):
        data_df = _pd.DataFrame(index=[0],
                                data={"label": "electricityResource",
                                      "building": 1,
                                      "active": 1,
                                      "to": "gridBus",
                                      "variable costs": 0.204,
                                      "CO2 impact": self.path,
                                      })
        result = stv.CommoditySourcesConverter.set_from_dataFrame(data_df, nr_of_buildings=1)
        expected_dict = {
            'data': {'id': 'electricityResource_B001', 'label': 'electricityResource', "building": 1,
                     "variable_costs": 0.204, "CO2_impact": self.path, 'color': '#1f78b4'},
            'classes': 'source',
        }

        # Flesh out test?
        self.assertDictEqual(result[0].get_nodal_infos(), expected_dict)


class TestBusesConverter(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.path = _pl.Path("..\\excels\\basic_example\\electricity_impact.csv")
        self.nodalData = stv.BusesConverter('gridBus', None, None, building=1, excess=1,
                                            excess_costs=0.024, active=True)
        self.nodalDataExtended = stv.BusesConverter('gridBus', None, None, building=1, excess=1,
                                                    excess_costs=0.024, active=True, shortage=True,
                                                    shortage_costs=0.012)

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'gridBus_B001', 'label': 'gridBus', "building": 1, "excess": 1, "excess_costs": 0.024,
                     'shortage': None, "shortage_costs": None, 'color': '#1f78b4'},
            'classes': 'bus',
        }
        self.assertDictEqual(result, expected_dict)

    def test_get_nodal_infos_extended(self):
        result = self.nodalDataExtended.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'gridBus_B001', 'label': 'gridBus', "building": 1, "excess": 1, "excess_costs": 0.024,
                     'shortage': True, "shortage_costs": 0.012, 'color': '#1f78b4'},
            'classes': 'bus',
        }
        self.assertDictEqual(result, expected_dict)

    def test_set_from_dataFrame(self):
        data_df = _pd.DataFrame(index=None,
                                data={"label": ["gridBus", "electricityProdBus", "electricityInBus", "shDemandBus"],
                                      "building": [1, 1, 1, 1],
                                      "active": [1, 1, 1, 1],
                                      "excess": [False, False, True, False],
                                      "excess costs": [None, None, -0.09, None],
                                      })
        result = stv.BusesConverter.set_from_dataFrame(data_df, nr_of_buildings=1)
        expected_dict = {
            'data': {'id': 'electricityInBus_B001', 'label': 'electricityInBus', "building": 1, "excess": 1,
                     "excess_costs": -0.09,
                     'shortage': None, "shortage_costs": None, 'color': '#1f78b4'},
            'classes': 'bus',
        }

        # Flesh out test?
        self.assertDictEqual(result[2].get_nodal_infos(), expected_dict)


class TestDemandConverter(_ut.TestCase):
    def setUp(self):
        """ TODO: Perhaps another case with a specific building model would make sense.
            Then, the appropriate edges need to be tested.
        """
        self.maxDiff = None
        self.path = _pl.Path("..\\excels\\basic_example\\electricity_impact.csv")
        self.nodalData = stv.DemandConverter('electricityDemand', None, None, building=1, fixed=1,
                                             nominal_value=1, active=True)
        self.nodalDataExtended = stv.DemandConverter('electricityDemand', None, None, building=1,
                                                     fixed=1, nominal_value=1, active=True, building_model=True)

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'electricityDemand_B001', 'label': 'electricityDemand', "building": 1, "fixed": 1,
                     "nominal_value": 1, 'building model': None, 'building model out': None, 'color': '#1f78b4'},
            'classes': 'demand',
        }
        self.assertDictEqual(result, expected_dict)

    def test_get_nodal_infos_extended(self):
        result = self.nodalDataExtended.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'electricityDemand_B001', 'label': 'electricityDemand', "building": 1, "fixed": 1,
                     "nominal_value": 1, 'building model': True, 'building model out': None, 'color': '#1f78b4'},
            'classes': 'demand',
        }
        self.assertDictEqual(result, expected_dict)

    def test_set_from_dataFrame(self):
        data_df = _pd.DataFrame(index=None,
                                data={"label": ["electricityDemand", "spaceHeatingDemand", "domesticHotWaterDemand"],
                                      "building": [1, 1, 1],
                                      "active": [1, 1, 1],
                                      "from": ["electricityInBus", "shDemandBus", "domesticHotWaterBus"],
                                      "fixed": [1, 1, 1],
                                      "nominal value": [1, 1, 1],
                                      "building model": [None, None, None],
                                      })
        result = stv.DemandConverter.set_from_dataFrame(data_df, nr_of_buildings=1)
        expected_dict = {
            'data': {'id': 'electricityDemand_B001', 'label': 'electricityDemand', "building": 1, "fixed": 1,
                     "nominal_value": 1, 'building model': None, 'building model out': None, 'color': '#1f78b4'},
            'classes': 'demand',
        }

        # Flesh out test?
        self.assertDictEqual(result[0].get_nodal_infos(), expected_dict)


class TestGridConnectionConverter(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.nodalData = stv.GridConnectionConverter('gridElectricity', "gridBus", "electricityInBus",
                                                     building=1, efficiency=1, active=True)

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'gridElectricity_B001', 'label': 'gridElectricity', "building": 1, "efficiency": 1,
                     'color': '#1f78b4'},
            "classes": "grid_connection"
        }
        self.assertDictEqual(result, expected_dict)

    def test_set_from_dataFrame(self):
        data_df = _pd.DataFrame(index=None,
                                data={"label": ["gridElectricity", "electricitySource", "producedElectricity"],
                                      "building": [1, 1, 1],
                                      "active": [1, 1, 1],
                                      "from": ["gridBus", "electricityProdBus", "electricityBus"],
                                      "to": ["electricityInBus", "electricityBus", "electricityInBus"],
                                      "efficiency": [1, 1, 1],
                                      })
        result = stv.GridConnectionConverter.set_from_dataFrame(data_df, nr_of_buildings=1)
        expected_dict = {
            'data': {'id': 'gridElectricity_B001', 'label': 'gridElectricity', "building": 1, "efficiency": 1,
                     'color': '#1f78b4'},
            "classes": "grid_connection"
        }

        # Flesh out test?
        self.assertDictEqual(result[0].get_nodal_infos(), expected_dict)


class TestTransformersConverter(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.nodalData = stv.TransformersConverter('HP', "electricityInBus", ["shSourceBus", "dhwStorageBus"],
                                                   building=1, efficiency=3.5, active=True,
                                                   capacity_DHW=500,
                                                   capacity_SH=500,
                                                   capacity_min=5,
                                                   lifetime=20,
                                                   maintenance=0.02,
                                                   installation=0,
                                                   planification=0,
                                                   invest_base=16679,
                                                   invest_cap=2152,
                                                   heat_impact=0,
                                                   elec_impact=0,
                                                   impact_cap=280.9
                                                   )

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'HP_B001', 'label': 'HP', "building": 1, "efficiency": 3.5, 'capacity_DHW': 500,
                     'capacity_SH': 500, 'capacity_el': None, 'capacity_min': 5, 'elec_impact': 0, 'heat_impact': 0,
                     'impact_cap': 280.9, 'installation': 0, 'invest_base': 16679, 'invest_cap': 2152,
                     'lifetime': 20, 'maintenance': 0.02, 'planification': 0, 'color': '#1f78b4'},
            "classes": "transformer"
        }
        self.assertDictEqual(result, expected_dict)

    def test_set_from_dataFrame(self):
        data_df = _pd.DataFrame(index=None,
                                data={"label": ["HP", "GWHP"],
                                      "building": [1, 1],
                                      "active": [1, 1],
                                      "from": ["electricityInBus", "electricityInBus"],
                                      "to": ["shSourceBus,dhwStorageBus", "shSourceBus,dhwStorageBus"],
                                      "efficiency": [3.5, 4.5],
                                      "capacity_DHW": [500, 500],
                                      "capacity_SH": [500, 500],
                                      "capacity_el": [None, None],
                                      "capacity_min": [5, 5],
                                      "lifetime": [20, 20],
                                      "maintenance": [0.02, 0.02],
                                      "installation": [0, 0],
                                      "planification": [0, 0],
                                      "invest_base": [16679, 22257],
                                      "invest_cap": [2152, 3052],
                                      "heat_impact": [0, 0],
                                      "elec_impact": [0, 0],
                                      "impact_cap": [280.9, 772]})

        result = stv.TransformersConverter.set_from_dataFrame(data_df, nr_of_buildings=1)
        expected_dict = {
            'data': {'id': 'HP_B001', 'label': 'HP', "building": 1, "efficiency": 3.5, 'capacity_DHW': 500,
                     'capacity_SH': 500, 'capacity_el': None, 'capacity_min': 5, 'elec_impact': 0, 'heat_impact': 0,
                     'impact_cap': 280.9, 'installation': 0, 'invest_base': 16679, 'invest_cap': 2152,
                     'lifetime': 20, 'maintenance': 0.02, 'planification': 0, 'color': '#1f78b4'},
            "classes": "transformer"
        }

        # Flesh out test?
        self.assertDictEqual(result[0].get_nodal_infos(), expected_dict)

        expected_dict_edge = {'data': {'source': 'HP_B001', 'target': 'shSourceBus_B001'},
                              'classes': 'SH', }

        self.assertDictEqual(result[0].get_edge_infos()[1], expected_dict_edge)


class TestStorageConverter(_ut.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.nodalData = stv.StoragesConverter('HP', "electricityInBus", ["shSourceBus", "dhwStorageBus"],
                                               building=1, active=True,
                                               efficiency_inflow=0.9,
                                               efficiency_outflow=0.86,
                                               initial_capacity=0,
                                               capacity_min=0,
                                               capacity_max=1000000,
                                               capacity_loss=0,
                                               lifetime=15,
                                               maintenance=0,
                                               installation=0,
                                               planification=0,
                                               invest_base=5138,
                                               invest_cap=981,
                                               heat_impact=0,
                                               elec_impact=0,
                                               impact_cap=28.66
                                               )

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'HP_B001', 'label': 'HP', "building": 1, "efficiency inflow": 0.9,
                     "efficiency outflow": 0.86,
                     'capacity loss': 0, 'capacity max': 1000000, 'capacity min': 0, 'elec_impact': 0, 'heat_impact': 0,
                     'impact_cap': 28.66, 'installation': 0, 'invest_base': 5138, 'invest_cap': 981,
                     'initial capacity': 0, 'lifetime': 15, 'maintenance': 0, 'planification': 0, 'color': '#1f78b4'},
            "classes": "storage"
        }
        self.assertDictEqual(result, expected_dict)

    def test_set_from_dataFrame(self):
        data_df = _pd.DataFrame(index=None,
                                data={"label": ["electricalStorage", "shStorage", "dhwStorage"],
                                      "building": [1, 1, 1],
                                      "active": [1, 1, 1],
                                      "from": ["electricityProdBus", "shSourceBus", "dhwStorageBus"],
                                      "to": ["electricityBus", "spaceHeatingBus", "domesticHotWaterBus"],
                                      "lifetime": [15, 20, 20],
                                      "maintenance": [0, 0, 0],
                                      "installation": [0, 0, 0],
                                      "planification": [0, 0, 0],
                                      "invest_base": [5138, 1092, 2132],
                                      "invest_cap": [981, 1.41, 6.88],
                                      "heat_impact": [0, 0, 0],
                                      "elec_impact": [0, 0, 0],
                                      "impact_cap": [28.66, 0.49, 0.49],
                                      "efficiency inflow": [0.9, None, None],
                                      "efficiency outflow": [0.86, None, None],
                                      "initial capacity": [0, 0, 0],
                                      "capacity min": [0, 0, 0],
                                      "capacity max": [1000000, 1000000, 1000000],
                                      "capacity loss": [0, None, None],
                                      }
                                )

        result = stv.StoragesConverter.set_from_dataFrame(data_df, nr_of_buildings=1)
        expected_dict = {
            'data': {'id': 'electricalStorage_B001', 'label': 'electricalStorage', "building": 1,
                     "efficiency inflow": 0.9, "efficiency outflow": 0.86,
                     'capacity loss': 0.0, 'capacity max': 1000000, 'capacity min': 0, 'elec_impact': 0,
                     'heat_impact': 0, 'impact_cap': 28.66, 'installation': 0, 'invest_base': 5138, 'invest_cap': 981.0,
                     'initial capacity': 0, 'lifetime': 15, 'maintenance': 0, 'planification': 0, 'color': '#1f78b4'},
            "classes": "storage"
        }

        # Flesh out test?
        self.assertDictEqual(result[0].get_nodal_infos(), expected_dict)

        expected_dict_edge = {'data': {'source': 'electricityProdBus_B001', 'target': 'electricalStorage_B001'},
                              'classes': 'electricity', }

        self.assertDictEqual(result[0].get_edge_infos()[0], expected_dict_edge)


class TestPVConverter(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.path = _pl.Path("..\\excels\\basic_example\\electricity_impact.csv")
        self.nodalData = stv.PVConverter('pv', None, 'electricityProdBus',
                                         active=True,
                                         building=1,
                                         peripheral_losses=0.05,
                                         latitude=47.49,
                                         longitude=7.59,
                                         tilt=30,
                                         azimuth=180,
                                         delta_temp_n=40,
                                         capacity_max=500,
                                         capacity_min=0.4,
                                         lifetime=30,
                                         maintenance=0.02,
                                         installation=0,
                                         planification=0,
                                         invest_base=17950,
                                         invest_cap=1103,
                                         heat_impact=0,
                                         elec_impact=0,
                                         impact_cap=1131,
                                         roof_area=400,
                                         zenith_angle=20)

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'pv_B001', 'label': 'pv', "building": 1,
                     'peripheral_losses': 0.05,
                     'latitude': 47.49,
                     'longitude': 7.59,
                     'tilt': 30,
                     'azimuth': 180,
                     'delta_temp_n': 40,
                     'capacity_max': 500,
                     'capacity_min': 0.4,
                     'lifetime': 30,
                     'maintenance': 0.02,
                     'installation': 0,
                     'planification': 0,
                     'invest_base': 17950,
                     'invest_cap': 1103,
                     'heat_impact': 0,
                     'elec_impact': 0,
                     'impact_cap': 1131,
                     'roof_area': 400,
                     'zenith_angle': 20,
                     'color': '#1f78b4', },
            'classes': 'solar',
        }
        self.assertDictEqual(result, expected_dict)


class TestSolarCollectorConverter(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.nodalData = stv.SolarCollectorConverter('solarCollector', 'electricityInBus', 'dhwStorageBus',
                                                     active=True,
                                                     building=1,
                                                     connect='solarConnectBus',
                                                     electrical_consumption=0.02,
                                                     peripheral_losses=0.05,
                                                     latitude=47.23,
                                                     longitude=8.34,
                                                     tilt=45,
                                                     azimuth=225,
                                                     eta_0=0.73,
                                                     a_1=1.7,
                                                     a_2=0.016,
                                                     temp_collector_inlet=20,
                                                     delta_temp_n=40,
                                                     capacity_max=100,
                                                     capacity_min=0,
                                                     lifetime=25,
                                                     maintenance=0.05,
                                                     installation=0.15,
                                                     planification=0.05,
                                                     invest_base=11034,
                                                     invest_cap=706.36,
                                                     heat_impact=0,
                                                     elec_impact=0,
                                                     impact_cap=0,
                                                     roof_area=400,
                                                     zenith_angle=20
                                                     )

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'solarCollector_B001', 'label': 'solarCollector', "building": 1,
                     'electrical_consumption': 0.02,
                     'peripheral_losses': 0.05,
                     'latitude': 47.23,
                     'longitude': 8.34,
                     'tilt': 45,
                     'azimuth': 225,
                     'eta_0': 0.73,
                     'a_1': 1.7,
                     'a_2': 0.016,
                     'temp_collector_inlet': 20,
                     'delta_temp_n': 40,
                     'capacity_max': 100,
                     'capacity_min': 0,
                     'lifetime': 25,
                     'maintenance': 0.05,
                     'installation': 0.15,
                     'planification': 0.05,
                     'invest_base': 11034,
                     'invest_cap': 706.36,
                     'heat_impact': 0,
                     'elec_impact': 0,
                     'impact_cap': 0,
                     'roof_area': 400,
                     'zenith_angle': 20,
                     'color': '#1f78b4', },
            'classes': 'solar',
        }
        self.assertDictEqual(result, expected_dict)


class TestSolarConverter(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_set_from_dataFrame(self):
        data_df = _pd.DataFrame(index=None,
                                data={
                                    "label": ['solarCollector', 'solarCollector', 'solarCollector', 'solarCollector',
                                              'pv', 'pv', 'pv', 'pv'],
                                    "building": [1, 2, 3, 4, 1, 2, 3, 4],
                                    "active": [1, 1, 1, 1, 1, 1, 1, 1],
                                    "from": ['electricityInBus', 'electricityInBus', 'electricityInBus',
                                             'electricityInBus', 'x', 'x', 'x', 'x'],
                                    "to": ['dhwStorageBus', 'dhwStorageBus', 'dhwStorageBus', 'dhwStorageBus',
                                           'electricityProdBus', 'electricityProdBus', 'electricityProdBus',
                                           'electricityProdBus'],
                                    "connect": ['solarConnectBus', 'solarConnectBus', 'solarConnectBus',
                                                'solarConnectBus', 'x', 'x', 'x', 'x'],
                                    "electrical_consumption": [0.02, 0.02, 0.02, 0.02, 'x', 'x', 'x', 'x'],
                                    "peripheral_losses": [0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05],
                                    "latitude": [47.23, 47.23, 47.23, 47.23, 47.23, 47.23, 47.23, 47.23],
                                    "longitude": [8.34, 8.34, 8.34, 8.34, 8.34, 8.34, 8.34, 8.34],
                                    "tilt": [45, 45, 45, 45, 45, 45, 45, 45],
                                    "azimuth": [225, 225, 225, 225, 225, 225, 225, 225],
                                    "efficiency": [None, None, None, None, 0.2, 0.2, 0.2, 0.2],
                                    "eta_0": [0.73, 0.73, 0.73, 0.73, 'x', 'x', 'x', 'x'],
                                    "a_1": [1.7, 1.7, 1.7, 1.7, 'x', 'x', 'x', 'x'],
                                    "a_2": [0.016, 0.016, 0.016, 0.016, 'x', 'x', 'x', 'x'],
                                    "temp_collector_inlet": [20, 20, 20, 20, 'x', 'x', 'x', 'x'],
                                    "delta_temp_n": [40, 40, 40, 40, 40, 40, 40, 40],
                                    "capacity_max": [100, 100, 100, 100, 100, 100, 100, 100],
                                    "capacity_min": [0, 0, 0, 0, 0.4, 0.4, 0.4, 0.4],
                                    "lifetime": [25, 25, 25, 25, 30, 30, 30, 30],
                                    "maintenance": [0.05, 0.05, 0.05, 0.05, 0, 0, 0, 0],
                                    "installation": [0.15, 0.15, 0.15, 0.15, 0, 0, 0, 0],
                                    "planification": [0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05],
                                    "invest_base": [11034, 11034, 11034, 11034, 17950, 17950, 17950, 17950],
                                    "invest_cap": [706.36, 706.36, 706.36, 706.36, 1103, 1103, 1103, 1103],
                                    "heat_impact": [0, 0, 0, 0, 0, 0, 0, 0],
                                    "elec_impact": [0, 0, 0, 0, 0, 0, 0, 0],
                                    "impact_cap": [0, 0, 0, 0, 0, 0, 0, 0],
                                    "roof_area": [400, 400, 400, 400, 400, 400, 400, 400],
                                    "zenith_angle": [20, 20, 20, 20, 20, 20, 20, 20],
                                })

        result = stv.SolarConverter.set_from_dataFrame(data_df, nr_of_buildings=4)

        # SolarCollector part
        expected_dict = {
            'data': {'id': 'solarCollector_B001', 'label': 'solarCollector', "building": 1,
                     'a_1': 1.7,
                     'a_2': 0.016,
                     'azimuth': 225,
                     'capacity_max': 100,
                     'capacity_min': 0.0,
                     'color': '#1f78b4',
                     'delta_temp_n': 40,
                     'elec_impact': 0,
                     'electrical_consumption': 0.02,
                     'eta_0': 0.73,
                     'heat_impact': 0,
                     'impact_cap': 0,
                     'installation': 0.15,
                     'invest_base': 11034,
                     'invest_cap': 706.36,
                     'latitude': 47.23,
                     'lifetime': 25,
                     'longitude': 8.34,
                     'maintenance': 0.05,
                     'peripheral_losses': 0.05,
                     'planification': 0.05,
                     'roof_area': 400,
                     'temp_collector_inlet': 20,
                     'tilt': 45,
                     'zenith_angle': 20,
                     },
            'classes': 'solar',
        }

        # Flesh out test?
        self.assertDictEqual(result[0].get_nodal_infos(), expected_dict)

        expected_dict_edge = {'data': {'source': 'electricityInBus_B001', 'target': 'solarCollector_B001'},
                              'classes': 'electricity', }

        self.assertDictEqual(result[0].get_edge_infos()[0], expected_dict_edge)

        expected_dict_edge = {'data': {'source': 'solarCollector_B002', 'target': 'dhwStorageBus_B002'},
                              'classes': 'DHW', }

        self.assertDictEqual(result[1].get_edge_infos()[1], expected_dict_edge)

        # pv part
        expected_dict = {
            'data': {'id': 'pv_B001', 'label': 'pv', "building": 1,
                     'azimuth': 225,
                     'capacity_max': 100,
                     'capacity_min': 0.4,
                     'color': '#1f78b4',
                     'delta_temp_n': 40,
                     'elec_impact': 0,
                     'heat_impact': 0,
                     'impact_cap': 0,
                     'installation': 0.0,
                     'invest_base': 17950,
                     'invest_cap': 1103.0,
                     'latitude': 47.23,
                     'lifetime': 30,
                     'longitude': 8.34,
                     'maintenance': 0.0,
                     'peripheral_losses': 0.05,
                     'planification': 0.05,
                     'roof_area': 400,
                     'tilt': 45,
                     'zenith_angle': 20,
                     },
            'classes': 'solar',
        }

        # Flesh out test?
        self.assertDictEqual(result[4].get_nodal_infos(), expected_dict)

        expected_dict_edge = {'data': {'source': 'pv_B001', 'target': 'electricityProdBus_B001'},
                              'classes': 'electricity', }

        self.assertDictEqual(result[4].get_edge_infos()[0], expected_dict_edge)

        expected_dict_edge = {'data': {'source': 'pv_B002', 'target': 'electricityProdBus_B002'},
                              'classes': 'electricity', }

        self.assertDictEqual(result[5].get_edge_infos()[0], expected_dict_edge)


class TestLinksConverter(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.nodalData = stv.LinksConverter('electricityLink', None, None, active=True, efficiency=0.9999,
                                            invest_base=0,
                                            invest_cap=0, investment=1, nr_of_buildings=1)

    def test_get_nodal_infos(self):
        result = self.nodalData.get_nodal_infos()
        expected_dict = {
            'data': {'id': 'electricityLink', 'label': 'electricityLink', 'efficiency': 0.9999, 'invest_base': 0,
                     'invest_cap': 0, 'investment': 1},
            'classes': 'link',
        }
        self.assertDictEqual(result, expected_dict)

    def test_set_from_dataFrame(self):
        data_df = _pd.DataFrame(index=None,
                                data={
                                    "label": ['electricityLink', 'shLink', 'dhwLink'],
                                    "active": [True, True, True],
                                    "efficiency": [0.9999, 0.9, 0.9],
                                    "invest_base": [0, 0, 0],
                                    "invest_cap": [0, 0, 0],
                                    "investment": [1, 1, 1],
                                })

        result = stv.LinksConverter.set_from_dataFrame(data_df, nr_of_buildings=4)
        expected_dict = {
            'data': {'id': 'electricityLink', 'label': 'electricityLink', 'efficiency': 0.9999, 'invest_base': 0,
                     'invest_cap': 0, 'investment': 1},
            'classes': 'link',
        }

        # Flesh out test?
        self.assertDictEqual(result[0].get_nodal_infos(), expected_dict)

        edges = result[0].get_edge_infos()
        assert len(edges) == 8

        self.assertDictEqual(edges[0], {'data': {'source': 'electricityBus_B001', 'target': 'electricityLink'},
                                        'classes': 'electricity', })
        self.assertDictEqual(edges[4], {'data': {'source': 'electricityLink', 'target': 'electricityInBus_B001'},
                                        'classes': 'electricity', })



import json
import os
from textwrap import dedent as d
import typing as _tp
import pathlib as _pl
import json as _json

import dash
from datetime import datetime
import plotly.express as px
from dash import html, dcc, Input, Output, State, callback, ctx, callback_context
from dash.exceptions import PreventUpdate
import dash_cytoscape as cyto
import matplotlib.pyplot as plt

import optihood.Visualizer.scenario_to_visualizer as stv
import optihood.Visualizer.convert_scenario as _cv

"""
    The dash technology restricts the passing of information to a list of dicts with limited elements (str, int, float).
    Thus, polymorphism is now working around the dictionary as follows:
    - Prepare dicts to be given to dash from polymorphic classes
    - Use same classes to read the dict again and produce the string needed for the UI.
    - These classes should be usable in different technology, which is likely required to move forward to a GUI.
"""


# TODO: better to overload input.

# TODO: extract style sheets and make methods to create them.
#       nodal_shapes = {'bus': 'rectangle'}
#       styles = get_styles(nodal_shapes)


def setup_cytoscape_app(graphData: _tp.Optional[_cv.EnergyNetworkGraphData] = None,
                        nodes: _tp.Optional[_tp.Dict[str, _tp.Dict[str, _tp.Union[str, float]]]] = None,
                        edges: _tp.Optional[_tp.Dict[str, _tp.Dict[str, _tp.Union[str, float]]]] = None,
                        node_layout_file: str = 'saved_layout.json',
                        layout_mode: str ='breadthfirst') -> dash.Dash:
    """
    This example comes directly from the Plotly homepage.
    http://dash.plotly.com/cytoscape/events
    """
    if not graphData and not nodes and not edges:
        raise TypeError('Though all inputs are optional, either graphData or both Nodes and Edges need to be defined.')

    if graphData:
        nodes = graphData.nodes
        edges = graphData.edges

    # enable svg export
    cyto.load_extra_layouts()
    app = dash.Dash('Optihood input Visualizer')
    styles = {
        'output': {
            'overflow-y': 'scroll',
            'overflow-wrap': 'break-word',
            'height': 'calc(100% - 25px)',
            'border': 'thin lightgrey solid'
        },
        'tab': {'height': 'calc(98vh - 115px)'}
    }

    default_stylesheet = [
        # selection colors
        {
            'selector': 'node',
            'style': {
                'label': 'data(label)',
                'background-color': 'data(color)',
                'font-size': '20px',
                'font-weight': 'bold'
            }
        },

        ## node shapes ##
        # ellipse
        # triangle
        # round-triangle
        # rectangle
        # round-rectangle
        # bottom-round-rectangle
        # cut-rectangle
        # barrel
        # rhomboid
        # right-rhomboid
        # diamond
        # round-diamond
        # pentagon
        # round-pentagon
        # hexagon
        # round-hexagon
        # concave-hexagon
        # heptagon
        # round-heptagon
        # octagon
        # round-octagon
        # star
        # tag
        # round-tag
        # vee
        {
            'selector': '.bus',
            'style': {
                "shape": 'rectangle',
            }
        },
        {
            'selector': '.demand',
            'style': {
                "shape": 'triangle',
            }
        },
        {
            'selector': '.transformer',
            'style': {
                "shape": 'right-rhomboid',
            }
        },
        {
            'selector': '.source',
            'style': {
                "shape": 'diamond',
            }
        },
        {
            'selector': '.grid_connection',
            'style': {
                "shape": 'hexagon',
            }
        },
        {
            'selector': '.storage',
            'style': {
                "shape": 'star',
            }
        },
        {
            'selector': '.solar',
            'style': {
                "shape": 'vee',
            }
        },
        {
            'selector': '.link',
            'style': {
                "shape": 'round-rectangle',
            }
        },

        # line styles
        {
            'selector': f'.{stv.EnergyTypes.electricity.value}',
            'style': {
                "line-style": 'dashed',
                'line-color': 'forestgreen',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.domestic_hot_water.value}',
            'style': {
                'line-color': 'tomato',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.space_heating.value}',
            'style': {
                'line-color': 'goldenrod',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.gas.value}',
            'style': {
                'curve-style': 'segments',
                'line-color': 'mediumpurple',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.unknown.value}',
            'style': {
                "line-style": 'dashed',
                'line-color': 'grey',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.low_temp_grid.value}',
            'style': {
                'line-color': 'dodgerblue',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.sh_temp_grid.value}',
            'style': {
                'line-color': 'goldenrod',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.med_temp_grid.value}',
            'style': {
                'line-color': 'orangered',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.dhw_temp_grid.value}',
            'style': {
                'line-color': 'tomato',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.biomass.value}',
            'style': {
                'line-color': 'yellowgreen',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.oil.value}',
            'style': {
                'line-color': 'royalblue',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.district.value}',
            'style': {
                'line-color': 'crimson',
            }
        },
        {
            'selector': f'.{stv.EnergyTypes.cooling.value}',
            'style': {
                'line-color': 'dodgerblue',
            }
        },

    ]

    app.layout = html.Div([
        cyto.Cytoscape(
            id='cytoscape-event-callbacks-3',
            layout={'name': layout_mode},  # 'cose', 'breadthfirst'
            elements=edges + nodes,
            stylesheet=default_stylesheet,
            style={'width': '100%', 'height': '640px'}
        ),
        html.Button("💾 Save Layout", id="save-button", n_clicks=0),
        html.Div(id="save-status", style={"marginTop": "10px", "color": "green"}),
        dcc.Markdown(id='cytoscape-selectedNodeData-markdown'),
    ])

    @callback(Output('cytoscape-selectedNodeData-markdown', 'children'),
              Input('cytoscape-event-callbacks-3', 'selectedNodeData'))
    def displaySelectedNodeData(data_list):
        if data_list is None:
            return "No nodes selected."

        NodalData = stv.scenario_data_factory(stv.ScenarioDataTypes.example)
        selected_nodes_list = [NodalData.read_nodal_infos(data) for data in data_list]
        return "You selected the nodes: " + "\n* ".join(selected_nodes_list)

    @callback(
        Output("save-status", "children"),
        Input("save-button", "n_clicks"),
        State("cytoscape-event-callbacks-3", "elements"),
        prevent_initial_call=True
    )
    def save_layout(_, elements):
        if not callback_context.triggered:
            raise PreventUpdate

        if elements:
            positions = {
                el['data']['id']: el['position']
                for el in elements if 'position' in el
            }
            with open(node_layout_file, "w") as f:
                json.dump(positions, f, indent=2)
            timestamp = datetime.now().strftime("%d-%m-%Y %H:%M:%S")
            return f"Layout saved successfully at {timestamp}. Location: {os.path.abspath(node_layout_file)}"

        return "No elements to save."

    return app


def setup_plotly_app(figure_handle: plt.Figure) -> dash.Dash:
    """
    This example comes directly from the Plotly homepage.
    https://community.plotly.com/t/use-hover-trace-as-input-for-callback/34390
    """
    app = dash.Dash('Optihood input Visualizer')
    styles = {
        'pre': {
            'border': 'thin lightgrey solid',
            'overflowX': 'scroll'
        }
    }

    app.layout = html.Div([
        dcc.Graph(
            id='basic-interactions',
            figure=figure_handle,
        ),

        html.Div(className='row', children=[
            html.Div([
                dcc.Markdown(d("""
                    **Hover Data**

                    Mouse over values in the graph.
                """)),
                html.Pre(id='hover-data', style=styles['pre'])
            ], className='three columns'),

            html.Div([
                dcc.Markdown(d("""
                    **Click Data**

                    Click on points in the graph.
                """)),
                html.Pre(id='click-data', style=styles['pre']),
            ], className='three columns'),

            html.Div([
                dcc.Markdown(d("""
                    **Selection Data**

                    Choose the lasso or rectangle tool in the graph's menu
                    bar and then select points in the graph.

                    Note that if `layout.clickmode = 'event+select'`, selection data also
                    accumulates (or un-accumulates) selected data if you hold down the shift
                    button while clicking.
                """)),
                html.Pre(id='selected-data', style=styles['pre']),
            ], className='three columns'),

            html.Div([
                dcc.Markdown(d("""
                    **Zoom and Relayout Data**

                    Click and drag on the graph to zoom or click on the zoom
                    buttons in the graph's menu bar.
                    Clicking on legend items will also fire
                    this event.
                """)),
                html.Pre(id='relayout-data', style=styles['pre']),
            ], className='three columns')
        ])
    ])

    @app.callback(
        Output('hover-data', 'children'),
        [Input('basic-interactions', 'hoverData')])
    def display_hover_data(hoverData):
        return json.dumps(hoverData, indent=2)

    @app.callback(
        Output('click-data', 'children'),
        [Input('basic-interactions', 'clickData')])
    def display_click_data(clickData):
        return json.dumps(clickData, indent=2)

    @app.callback(
        Output('selected-data', 'children'),
        [Input('basic-interactions', 'selectedData')])
    def display_selected_data(selectedData):
        return json.dumps(selectedData, indent=2)

    @app.callback(
        Output('relayout-data', 'children'),
        [Input('basic-interactions', 'relayoutData')])
    def display_relayout_data(relayoutData):
        return json.dumps(relayoutData, indent=2)

    return app


def run_fig_visualizer(figure_handle: plt.Figure) -> None:
    app = setup_plotly_app(figure_handle)
    app.run_server(debug=True)


def run_cytoscape_visualizer(
        graphData: _tp.Optional[_cv.EnergyNetworkGraphData] = None,
        nodes: _tp.Optional[_tp.Dict[str, _tp.Dict[str, _tp.Union[str, float]]]] = None,
        edges: _tp.Optional[_tp.Dict[str, _tp.Dict[str, _tp.Union[str, float]]]] = None,
        node_layout_file: str = 'saved_layout.json',
        layout_mode: _tp.Literal["breadthfirst", "circle", "cose", "grid", "random"] = 'breadthfirst'
        ) -> None:
    """
    Method to visualize an energy network.
    Takes either a EnergyNetworkGraphData object, or the combination of both nodes and edges dictionaries.
    
    This visualizer runs in a web browser.
    There you can move the nodes to a better position and save the positions to the "node_layout_file".

    Parameters
    ----------
    graphData:
        optional EnergyNetworkGraphData object with nodes and edges.

    nodes:
        optional dictionary of "nodes" in the network. These correspond to the components, links, buses, etc.

    edges:
        optional dictionary of "edges" in the network. These correspond to flows between nodes and connect them.

    node_layout_file:
        file where the node positions should be stored.

    layout_mode:
        When no node_layout_file is available yet, the nodes will be distributed automatically using this algorithm.
        Options are: "breadthfirst", "circle", "cose", "grid" and "random".
    """
    if node_layout_file:
        node_layout_path = _pl.Path(node_layout_file)
        if node_layout_path.exists():
            with node_layout_path.open("r") as f:
                saved_positions = _json.load(f)

            for node in graphData.nodes:
                node_id = node['data']['id']
                if node_id in saved_positions:
                    node['position'] = saved_positions[node_id]

            layout_mode = 'preset'
    app = setup_cytoscape_app(graphData, nodes, edges, node_layout_file, layout_mode)
    app.run(debug=True)


if __name__ == '__main__':

    version = "cytoscape"
    if version == "plotly":
        df = px.data.iris()
        fig = px.scatter_matrix(df[['sepal_length', 'sepal_width']])
        run_fig_visualizer(fig)
    elif version == "cytoscape":
        NodalData = stv.scenario_data_factory(stv.ScenarioDataTypes.example)
        energy_type = stv.EnergyTypes.electricity
        nodal_data_list = [
            NodalData('la', 'Los_Angeles', 'van', 'chi', energy_type, True, 34.03, -118.25),
            NodalData('nyc', 'New_York', 'to', 'bos', energy_type, True, 40.71, -74),
            NodalData('to', 'Toronto', 'hou', ['hou', 'nyc', 'mtl'], energy_type, True, 43.65, -79.38),
            NodalData('mtl', 'Montreal', 'to', 'bos', energy_type, True, 45.50, -73.57),
            NodalData('van', 'Vancouver', None, 'la', energy_type, True, 49.28, -123.12),
            NodalData('chi', 'Chicago', 'la', 'hou', energy_type, True, 41.88, -87.63),
            NodalData('bos', 'Boston', "nyc", None, energy_type, True, 42.36, -71.06),
            NodalData('hou', 'Houston', "chi", 'chi', energy_type, True, 29.76, -95.37)
        ]
        nodes_dict = []
        edges_dict = []
        for i, nodalData in enumerate(nodal_data_list):
            nodes_dict.append(nodalData.get_nodal_infos())
            edges_dict += nodalData.get_edge_infos()

        run_cytoscape_visualizer(nodes_dict, edges_dict)
